# compatibility layer for offcscreen canvas and regular canvas
import sys
import os
from IPython.display import display

# this is very usefull for testing compatibility with offscreen canvas
IPYCANVAS_DISABLE_OFFSCREEN_CANVAS =  bool(int(os.environ.get('IPYCANVAS_DISABLE_OFFSCREEN_CANVAS', '0')))

import sys
is_emscripten = sys.platform.startswith("emscripten")

# has pyjs
has_pyjs = True
try:
    import pyjs
except ImportError:
    has_pyjs = False

if (is_emscripten and has_pyjs) and not IPYCANVAS_DISABLE_OFFSCREEN_CANVAS:
    from .offscreen_canvas.offscreen_canvas import OffscreenCanvas as Canvas
else:
    from .canvas import Canvas as CanvasBase

    class Canvas(CanvasBase):
        """Compatibility layer for offscreen canvas and regular canvas."""
        
        def initialize():
            """
                After the canvas has been displayed, we need to call this method to initialize the canvas.
                This method can only be used when the display method has been called in a **different cell**.

                # Cell1:
                ```python
                from ipycanvas.compat import Canvas
                from IPython.display import display
                canvas = Canvas(width=800, height=600)
                display(canvas)
                ```

                # Cell2:
                ```python
                canvas.initialize()
                ```

                For more information, see the `async_initialize` method.

            """
        
        async def async_initialize(self):
            """
        
                If we want to use the canvas in the same cell where it 
                was created **and displayed** we need to call this async function.
                While this sounds a bit counterintuitive, it is necessary because the
                offscreen canvas is created as regular canvas in the main-thread,
                and transfered to and recieved by the worker-thread. This transfering mechanism (in particular the
                receiving part) would be blocked by the cell execution.
                To get a chance to  receive the canvas in the worker-thread (ie where
                the kernel is running), we need to do run some asyc code (with some sleeping in between)
                Note that for an ordinary canvas, `async_initialize` is a no-op.

            from ipycanvas.compat import Canvas
            from IPython.display import display
            canvas = Canvas(width=800, height=600)
            display(canvas)
            await canvas.async_initialize()

            # canvas is now ready to use
            canvas.fill_style = 'red'


            """
        
        async def display(self):
            """ shorthand for displaying the canvas and then initializing it.
            See `async_initialize` for more information.
            """

            display(self)



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Martin Renou.
# Distributed under the terms of the Modified BSD License.

from .canvas import (
    Path2D,
    Canvas,
    RoughCanvas,
    MultiCanvas,
    MultiRoughCanvas,
    hold_canvas,
)  # noqa
from ._version import __version__  # noqa


def _jupyter_nbextension_paths():
    return [
        {
            "section": "notebook",
            "src": "nbextension/static",
            "dest": "ipycanvas",
            "require": "ipycanvas/extension",
        }
    ]


def _jupyter_labextension_paths():
    return [
        {
            "src": "labextension",
            "dest": "ipycanvas",
        }
    ]



"""Binary module."""

from io import BytesIO

from PIL import Image as PILImage

import numpy as np

try:
    import orjson

    ORJSON_AVAILABLE = True
except ImportError:
    import json

    ORJSON_AVAILABLE = False


def image_bytes_to_array(im_bytes):
    """Turn raw image bytes into a NumPy array."""
    im_file = BytesIO(im_bytes)

    im = PILImage.open(im_file)

    return np.array(im)


def binary_image(ar, quality=75):
    f = BytesIO()
    if len(ar.shape) == 2:
        filetype = "PNG"
    elif ar.shape[2] == 3:
        filetype = "JPEG"
    else:
        filetype = "PNG"
    PILImage.fromarray(ar.astype(np.uint8)).save(f, filetype, quality=quality)
    return f.getvalue()


def array_to_binary(ar):
    """Turn a NumPy array into a binary buffer."""
    # Unsupported int64 array JavaScript side
    if ar.dtype == np.int64:
        ar = ar.astype(np.int32)

    # Unsupported float16 array JavaScript side
    if ar.dtype == np.float16:
        ar = ar.astype(np.float32)

    # make sure it's contiguous
    if not ar.flags["C_CONTIGUOUS"]:
        ar = np.ascontiguousarray(ar)

    return {"shape": ar.shape, "dtype": str(ar.dtype)}, memoryview(ar)


def populate_args(arg, args, buffers):
    if isinstance(arg, (list, np.ndarray)):
        arg_metadata, arg_buffer = array_to_binary(np.asarray(arg))
        arg_metadata["idx"] = len(buffers)

        args.append(arg_metadata)
        buffers.append(arg_buffer)
    else:
        args.append(arg)


def commands_to_buffer(commands):
    # Turn the commands list into a binary buffer
    if ORJSON_AVAILABLE:
        return array_to_binary(
            np.frombuffer(
                bytes(orjson.dumps(commands, option=orjson.OPT_SERIALIZE_NUMPY)),
                dtype=np.uint8,
            )
        )
    else:
        return array_to_binary(
            np.frombuffer(bytes(json.dumps(commands), encoding="utf8"), dtype=np.uint8)
        )



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) Martin Renou.
# Distributed under the terms of the Modified BSD License.

import warnings
from contextlib import contextmanager

import numpy as np

from traitlets import (
    Bool,
    Bytes,
    CInt,
    Enum,
    Float,
    Instance,
    List,
    Unicode,
    TraitError,
    Union,
)

from ipywidgets import (
    CallbackDispatcher,
    Color,
    DOMWidget,
    Image,
    Widget,
    widget_serialization,
)
from ipywidgets.widgets.trait_types import (
    bytes_serialization,
    _color_names,
    _color_hex_re,
    _color_hexa_re,
    _color_rgbhsl_re,
)

from ._frontend import module_name, module_version

from .utils import binary_image, populate_args, image_bytes_to_array, commands_to_buffer

_CMD_LIST = [
    "fillRect",
    "strokeRect",
    "fillRects",
    "strokeRects",
    "clearRect",
    "fillArc",
    "fillCircle",
    "strokeArc",
    "strokeCircle",
    "fillArcs",
    "strokeArcs",
    "fillCircles",
    "strokeCircles",
    "strokeLine",
    "beginPath",
    "closePath",
    "stroke",
    "strokePath",
    "fillPath",
    "fill",
    "moveTo",
    "lineTo",
    "rect",
    "arc",
    "ellipse",
    "arcTo",
    "quadraticCurveTo",
    "bezierCurveTo",
    "fillText",
    "strokeText",
    "setLineDash",
    "drawImage",
    "putImageData",
    "clip",
    "save",
    "restore",
    "translate",
    "rotate",
    "scale",
    "transform",
    "setTransform",
    "resetTransform",
    "set",
    "clear",
    "sleep",
    "fillPolygon",
    "strokePolygon",
    "strokeLines",
    "fillPolygons",
    "strokePolygons",
    "strokeLineSegments",
    "fillStyledRects",
    "strokeStyledRects",
    "fillStyledCircles",
    "strokeStyledCircles",
    "fillStyledArcs",
    "strokeStyledArcs",
    "fillStyledPolygons",
    "strokeStyledPolygons",
    "strokeStyledLineSegments",
    "switchCanvas",
]
COMMANDS = {v: i for i, v in enumerate(_CMD_LIST)}


# Traitlets does not allow validating without creating a trait class, so we need this
def _validate_color(value):
    if isinstance(value, str):
        if (
            value.lower() in _color_names
            or _color_hex_re.match(value)
            or _color_hexa_re.match(value)
            or _color_rgbhsl_re.match(value)
        ):
            return value
    raise TraitError("{} is not a valid HTML Color".format(value))


def _validate_number(value, min_val, max_val):
    try:
        number = float(value)

        if number >= min_val and number <= max_val:
            return number
    except ValueError:
        raise TraitError("{} is not a number".format(value))
    raise TraitError("{} is not in the range [{}, {}]".format(value, min_val, max_val))


def _serialize_list_of_polygons_or_linestrokes(
    points, points_per_item, item_name, min_elements
):
    if isinstance(points, list):
        if points_per_item is not None:
            raise RuntimeError("when points are a list, points_per_item must be None")
        points_per_item = []
        np_polygons = []
        for i, polygon_points in enumerate(points):
            polygon_points = np.require(polygon_points, requirements=["C"])
            if polygon_points.shape[1] != 2:
                raise RuntimeError(
                    f"item {i} in points have wrong shape: `{polygon_points.shape}` but must be of type (n,2)"
                )
            points_per_item.append(polygon_points.shape[0])
            np_polygons.append(polygon_points.ravel())

        num_polygons = len(points)
        flat_points = np.concatenate(np_polygons)
        points_per_item = np.array(points_per_item)

    elif isinstance(points, np.ndarray):
        points = np.require(points, requirements=["C"])
        shape = points.shape
        ndim = points.ndim

        if ndim <= 2:
            if points_per_item is None:
                raise RuntimeError(
                    "when points are given as a 1d / 2d array, points_per_item must not be None"
                )
            if ndim == 1:
                flat_points = points
            elif ndim == 2:
                if shape[1] != 2:
                    raise RuntimeError(
                        f"points have wrong shape: `{shape}`. When points are given as a 2D array the shape must be of (n,2)"
                    )
                flat_points = points.ravel()
            num_polygons = len(points_per_item)
        elif ndim == 3:
            if points_per_item is not None:
                raise RuntimeError(
                    "when points are a list, points_per_item must be None"
                )
            if shape[2] != 2:
                raise RuntimeError(
                    f"Points have wrong shape: `{shape}`: When points are given as a 3D array the shape must be of (n_{item_name}, n_points_per_{item_name}, 2)"
                )
            if shape[1] < min_elements:
                raise RuntimeError(
                    f"Points have wrong shape: `{shape}`: when points are given as a 3D array the shape must be of (n_{item_name}, n_points_per_{item_name}, 2) and n_points_per_{item_name} must be >= {min_elements} "
                )
            flat_points = points.ravel()
            num_polygons = shape[0]
            points_per_item = shape[1]
        else:
            raise RuntimeError("ndarray must have ndim <= 3")
    else:
        raise RuntimeError("points must be a list or an ndarray")
    return num_polygons, flat_points, points_per_item


class _CanvasManager(Widget):
    """Private Canvas manager."""

    _model_module = Unicode(module_name).tag(sync=True)
    _model_module_version = Unicode(module_version).tag(sync=True)

    _model_name = Unicode("CanvasManagerModel").tag(sync=True)

    def __init__(self, *args, **kwargs):
        self._caching = kwargs.get("caching", False)
        self._commands_cache = []
        self._buffers_cache = []
        self._current_canvas = None

        super(_CanvasManager, self).__init__()

    def send_draw_command(self, canvas, name, args=[], buffers=[]):
        while len(args) and args[len(args) - 1] is None:
            args.pop()
        self.send_command(canvas, [name, args, len(buffers)], buffers)

    def send_command(self, canvas, command, buffers=[]):
        if self._caching:
            if self._current_canvas is not canvas:
                self._commands_cache.append(
                    [
                        COMMANDS["switchCanvas"],
                        [widget_serialization["to_json"](canvas, None)],
                    ]
                )
                self._current_canvas = canvas
            self._commands_cache.append(command)
            self._buffers_cache += buffers

            return

        # TODO Send the switch and the message in one batch?
        if self._current_canvas is not canvas:
            self._send_custom(
                [
                    COMMANDS["switchCanvas"],
                    [widget_serialization["to_json"](canvas, None)],
                ]
            )
            self._current_canvas = canvas

        self._send_custom(command, buffers)

    def flush(self):
        """Flush all the cached commands and clear the cache."""
        if not self._caching or not len(self._commands_cache):
            return

        self._send_custom(self._commands_cache, self._buffers_cache)

        self._commands_cache = []
        self._buffers_cache = []

    def _send_custom(self, command, buffers=[]):
        metadata, command_buffer = commands_to_buffer(command)
        self.send(metadata, buffers=[command_buffer] + buffers)


# Main canvas manager
_CANVAS_MANAGER = _CanvasManager()


class Path2D(Widget):
    """Create a Path2D.

    Args:
        value (str): The path value, e.g. "M10 10 h 80 v 80 h -80 Z"
    """

    _model_module = Unicode(module_name).tag(sync=True)
    _model_module_version = Unicode(module_version).tag(sync=True)

    _model_name = Unicode("Path2DModel").tag(sync=True)

    value = Unicode(allow_none=False, read_only=True).tag(sync=True)

    def __init__(self, value):
        """Create a Path2D object given the path string."""
        self.set_trait("value", value)

        super(Path2D, self).__init__()


class Pattern(Widget):
    """Create a Pattern.

    Args:
        image (Canvas or MultiCanvas or ipywidgets.Image): The source to be used as the pattern's image
        repetition (str): A string indicating how to repeat the pattern's image, can be "repeat" (both directions), "repeat-x" (horizontal only), "repeat-y" (vertical only), "no-repeat" (neither direction)
    """

    _model_module = Unicode(module_name).tag(sync=True)
    _model_module_version = Unicode(module_version).tag(sync=True)

    _model_name = Unicode("PatternModel").tag(sync=True)

    image = Union(
        (
            Instance(Image),
            Instance("ipycanvas.Canvas"),
            Instance("ipycanvas.MultiCanvas"),
        ),
        allow_none=False,
        read_only=True,
    ).tag(sync=True, **widget_serialization)
    repetition = Enum(
        ["repeat", "repeat-x", "repeat-y", "no-repeat"],
        allow_none=False,
        read_only=True,
    ).tag(sync=True)

    def __init__(self, image, repetition="repeat"):
        """Create a Pattern object given the image and the type of repetition."""
        self.set_trait("image", image)
        self.set_trait("repetition", repetition)

        super(Pattern, self).__init__()

    def _ipython_display_(self, *args, **kwargs):
        return self.image._ipython_display_(*args, **kwargs)


class _CanvasGradient(Widget):
    _model_module = Unicode(module_name).tag(sync=True)
    _model_module_version = Unicode(module_version).tag(sync=True)

    x0 = Float(allow_none=False, read_only=True).tag(sync=True)
    y0 = Float(allow_none=False, read_only=True).tag(sync=True)
    x1 = Float(allow_none=False, read_only=True).tag(sync=True)
    y1 = Float(allow_none=False, read_only=True).tag(sync=True)

    color_stops = List(allow_none=False, read_only=True).tag(sync=True)

    def __init__(self, x0, y0, x1, y1, color_stops):
        self.set_trait("x0", x0)
        self.set_trait("y0", y0)
        self.set_trait("x1", x1)
        self.set_trait("y1", y1)

        for color_stop in color_stops:
            _validate_number(color_stop[0], 0, 1)
            _validate_color(color_stop[1])
        self.set_trait("color_stops", color_stops)

        super(_CanvasGradient, self).__init__()


class LinearGradient(_CanvasGradient):
    """Create a LinearGradient."""

    _model_name = Unicode("LinearGradientModel").tag(sync=True)

    def __init__(self, x0, y0, x1, y1, color_stops):
        """Create a LinearGradient object given the start point, end point and color stops.

        Args:
            x0 (float): The x-axis coordinate of the start point.
            y0 (float): The y-axis coordinate of the start point.
            x1 (float): The x-axis coordinate of the end point.
            y1 (float): The y-axis coordinate of the end point.
            color_stops (list): The list of color stop tuples (offset, color) defining the gradient.
        """
        super(LinearGradient, self).__init__(x0, y0, x1, y1, color_stops)


class RadialGradient(_CanvasGradient):
    """Create a RadialGradient."""

    _model_name = Unicode("RadialGradientModel").tag(sync=True)

    r0 = Float(allow_none=False, read_only=True).tag(sync=True)
    r1 = Float(allow_none=False, read_only=True).tag(sync=True)

    def __init__(self, x0, y0, r0, x1, y1, r1, color_stops):
        """Create a RadialGradient object given the start circle, end circle and color stops.

        Args:
            x0 (float): The x-axis coordinate of the start circle.
            y0 (float): The y-axis coordinate of the start circle.
            r0 (float): The radius of the start circle.
            x1 (float): The x-axis coordinate of the end circle.
            y1 (float): The y-axis coordinate of the end circle.
            r1 (float): The radius of the end circle.
            color_stops (list): The list of color stop tuples (offset, color) defining the gradient.
        """
        _validate_number(r0, 0, float("inf"))
        _validate_number(r1, 0, float("inf"))

        self.set_trait("r0", r0)
        self.set_trait("r1", r1)

        super(RadialGradient, self).__init__(x0, y0, x1, y1, color_stops)


class _CanvasBase(DOMWidget):
    _model_module = Unicode(module_name).tag(sync=True)
    _model_module_version = Unicode(module_version).tag(sync=True)
    _view_module = Unicode(module_name).tag(sync=True)
    _view_module_version = Unicode(module_version).tag(sync=True)

    _canvas_manager = Instance(_CanvasManager, default_value=_CANVAS_MANAGER).tag(
        sync=True, **widget_serialization
    )

    width = CInt(700).tag(sync=True)
    height = CInt(500).tag(sync=True)

    #: (bool) Specifies if the image should be synchronized from front-end to Python back-end
    sync_image_data = Bool(False).tag(sync=True)

    #: (bytes) Current image data as bytes (PNG encoded). It is ``None`` by default and will not be
    #: updated if ``sync_image_data`` is ``False``.
    image_data = Bytes(default_value=None, allow_none=True, read_only=True).tag(
        sync=True, **bytes_serialization
    )

    def to_file(self, filename):
        """Save the current Canvas image to a PNG file.

        This will raise an exception if there is no image to save (_e.g._ if ``image_data`` is ``None``).
        """
        if self.image_data is None:
            raise RuntimeError(
                "No image data to save, please be sure that ``sync_image_data`` is set to True"
            )
        if not filename.endswith(".png") and not filename.endswith(".PNG"):
            raise RuntimeError("Can only save to a PNG file")

        with open(filename, "wb") as fobj:
            fobj.write(self.image_data)

    def get_image_data(self, x=0, y=0, width=None, height=None):
        """Return a NumPy array representing the underlying pixel data for a specified portion of the canvas.

        This will throw an error if there is no ``image_data`` to retrieve, this happens when nothing was drawn yet or
        when the ``sync_image_data`` attribute is not set to ``True``.
        The returned value is a NumPy array containing the image data for the rectangle of the canvas specified. The
        coordinates of the rectangle's top-left corner are (``x``, ``y``), while the coordinates of the bottom corner
        are (``x + width``, ``y + height``).
        """
        if self.image_data is None:
            raise RuntimeError(
                "No image data, please be sure that ``sync_image_data`` is set to True"
            )

        x = int(x)
        y = int(y)

        if width is None:
            width = self.width - x
        if height is None:
            height = self.height - y

        width = int(width)
        height = int(height)

        image_data = image_bytes_to_array(self.image_data)
        return image_data[y : y + height, x : x + width]


class Canvas(_CanvasBase):
    """Create a Canvas widget.

    Args:
        width (int): The width (in pixels) of the canvas
        height (int): The height (in pixels) of the canvas
    """

    _model_name = Unicode("CanvasModel").tag(sync=True)
    _view_name = Unicode("CanvasView").tag(sync=True)

    _send_client_ready_event = Bool(True).tag(sync=True)

    #: (valid HTML color or Gradient or Pattern) The color for filling rectangles and paths. Default to ``'black'``.
    fill_style = Union(
        (Color(), Instance(_CanvasGradient), Instance(Pattern)), default_value="black"
    )

    #: (valid HTML color or Gradient or Pattern) The color for rectangles and paths stroke. Default to ``'black'``.
    stroke_style = Union(
        (Color(), Instance(_CanvasGradient), Instance(Pattern)), default_value="black"
    )

    #: (float) Transparency level. Default to ``1.0``.
    global_alpha = Float(1.0)

    #: (str) Font for the text rendering. Default to ``'12px serif'``.
    font = Unicode("12px serif")

    #: (str) Text alignment, possible values are ``'start'``, ``'end'``, ``'left'``, ``'right'``, and ``'center'``.
    #: Default to ``'start'``.
    text_align = Enum(
        ["start", "end", "left", "right", "center"], default_value="start"
    )

    #: (str) Text baseline, possible values are ``'top'``, ``'hanging'``, ``'middle'``, ``'alphabetic'``, ``'ideographic'``
    #: and ``'bottom'``.
    #: Default to ``'alphabetic'``.
    text_baseline = Enum(
        ["top", "hanging", "middle", "alphabetic", "ideographic", "bottom"],
        default_value="alphabetic",
    )

    #: (str) Text direction, possible values are ``'ltr'``, ``'rtl'``, and ``'inherit'``.
    #: Default to ``'inherit'``.
    direction = Enum(["ltr", "rtl", "inherit"], default_value="inherit")

    #: (str) Global composite operation, possible values are listed below:
    #: https://developer.mozilla.org/en-US/docs/Web/API/Canvas_API/Tutorial/Compositing#globalCompositeOperation
    global_composite_operation = Enum(
        [
            "source-over",
            "source-in",
            "source-out",
            "source-atop",
            "destination-over",
            "destination-in",
            "destination-out",
            "destination-atop",
            "lighter",
            "copy",
            "xor",
            "multiply",
            "screen",
            "overlay",
            "darken",
            "lighten",
            "color-dodge",
            "color-burn",
            "hard-light",
            "soft-light",
            "difference",
            "exclusion",
            "hue",
            "saturation",
            "color",
            "luminosity",
        ],
        default_value="source-over",
    )

    #: (float) Indicates the horizontal distance the shadow should extend from the object.
    #: This value isn't affected by the transformation matrix. The default is 0.
    shadow_offset_x = Float(0.0)

    #: (float) Indicates the vertical distance the shadow should extend from the object.
    #: This value isn't affected by the transformation matrix. The default is 0.
    shadow_offset_y = Float(0.0)

    #: (float) Indicates the size of the blurring effect; this value doesn't correspond to a number of pixels
    #: and is not affected by the current transformation matrix. The default value is 0.
    shadow_blur = Float(0.0)

    #: (valid HTML color) A standard CSS color value indicating the color of the shadow effect; by default,
    #: it is fully-transparent black.
    shadow_color = Color("rgba(0, 0, 0, 0)")

    #: (float) Sets the width of lines drawn in the future, must be a positive number. Default to ``1.0``.
    line_width = Float(1.0)

    #: (str) Sets the appearance of the ends of lines, possible values are ``'butt'``, ``'round'`` and ``'square'``.
    #: Default to ``'butt'``.
    line_cap = Enum(["butt", "round", "square"], default_value="butt")

    #: (str) Sets the appearance of the "corners" where lines meet, possible values are ``'round'``, ``'bevel'`` and ``'miter'``.
    #: Default to ``'miter'``
    line_join = Enum(["round", "bevel", "miter"], default_value="miter")

    #: (float) Establishes a limit on the miter when two lines join at a sharp angle, to let you control how thick
    #: the junction becomes. Default to ``10.``.
    miter_limit = Float(10.0)

    #: (str) Filter effects such as blurring and grayscaling. It is similar to the CSS filter property and accepts the same values.
    #: This property has no effect on Safari, see https://bugs.webkit.org/show_bug.cgi?id=198416
    filter = Unicode("none")

    #: (bool) Indicates if scaled images are smoothed. Default to True
    image_smoothing_enabled = Bool(True)

    _line_dash = List()

    #: (float) Specifies where to start a dash array on a line. Default is ``0.``.
    line_dash_offset = Float(0.0)

    _client_ready_callbacks = Instance(CallbackDispatcher, ())

    _mouse_wheel_callbacks = Instance(CallbackDispatcher, ())
    _mouse_move_callbacks = Instance(CallbackDispatcher, ())
    _mouse_down_callbacks = Instance(CallbackDispatcher, ())
    _mouse_up_callbacks = Instance(CallbackDispatcher, ())
    _mouse_out_callbacks = Instance(CallbackDispatcher, ())

    _touch_start_callbacks = Instance(CallbackDispatcher, ())
    _touch_end_callbacks = Instance(CallbackDispatcher, ())
    _touch_move_callbacks = Instance(CallbackDispatcher, ())
    _touch_cancel_callbacks = Instance(CallbackDispatcher, ())

    _key_down_callbacks = Instance(CallbackDispatcher, ())

    ATTRS = {
        "fill_style": 0,
        "stroke_style": 1,
        "global_alpha": 2,
        "font": 3,
        "text_align": 4,
        "text_baseline": 5,
        "direction": 6,
        "global_composite_operation": 7,
        "line_width": 8,
        "line_cap": 9,
        "line_join": 10,
        "miter_limit": 11,
        "line_dash_offset": 12,
        "shadow_offset_x": 13,
        "shadow_offset_y": 14,
        "shadow_blur": 15,
        "shadow_color": 16,
        "filter": 17,
        "image_smoothing_enabled": 18,
    }

    def __init__(self, *args, **kwargs):
        """Create a Canvas widget."""
        super(Canvas, self).__init__(*args, **kwargs)

        if "caching" in kwargs:
            self._canvas_manager._caching = kwargs["caching"]

            warnings.warn(
                "caching is deprecated and will be removed in a future release, please use hold_canvas() instead.",
                DeprecationWarning,
            )

        self.on_msg(self._handle_frontend_event)

    def sleep(self, time):
        """Make the Canvas sleep for `time` milliseconds."""
        self._canvas_manager.send_draw_command(self, COMMANDS["sleep"], [time])

    # Gradient methods
    def create_linear_gradient(self, x0, y0, x1, y1, color_stops):
        """Create a LinearGradient object given the start point, end point, and color stops.

        Args:
            x0 (float): The x-axis coordinate of the start point.
            y0 (float): The y-axis coordinate of the start point.
            x1 (float): The x-axis coordinate of the end point.
            y1 (float): The y-axis coordinate of the end point.
            color_stops (list): The list of color stop tuples (offset, color) defining the gradient.
        """
        return LinearGradient(x0, y0, x1, y1, color_stops)

    def create_radial_gradient(self, x0, y0, r0, x1, y1, r1, color_stops):
        """Create a RadialGradient object given the start circle, end circle and color stops.

        Args:
            x0 (float): The x-axis coordinate of the start circle.
            y0 (float): The y-axis coordinate of the start circle.
            r0 (float): The radius of the start circle.
            x1 (float): The x-axis coordinate of the end circle.
            y1 (float): The y-axis coordinate of the end circle.
            r1 (float): The radius of the end circle.
            color_stops (list): The list of color stop tuples (offset, color) defining the gradient.
        """
        return RadialGradient(x0, y0, r0, x1, y1, r1, color_stops)

    # Pattern method
    def create_pattern(self, image, repetition="repeat"):
        """Create a Pattern.

        Args:
            image (Canvas or MultiCanvas or ipywidgets.Image): The source to be used as the pattern's image
            repetition (str): A string indicating how to repeat the pattern's image, can be "repeat" (both directions), "repeat-x" (horizontal only), "repeat-y" (vertical only), "no-repeat" (neither direction)
        """
        return Pattern(image, repetition)

    # Rectangles methods
    def fill_rect(self, x, y, width, height=None):
        """Draw a filled rectangle of size ``(width, height)`` at the ``(x, y)`` position."""
        if height is None:
            height = width

        self._canvas_manager.send_draw_command(
            self, COMMANDS["fillRect"], [x, y, width, height]
        )

    def stroke_rect(self, x, y, width, height=None):
        """Draw a rectangular outline of size ``(width, height)`` at the ``(x, y)`` position."""
        if height is None:
            height = width

        self._canvas_manager.send_draw_command(
            self, COMMANDS["strokeRect"], [x, y, width, height]
        )

    def fill_rects(self, x, y, width, height=None):
        """Draw filled rectangles of sizes ``(width, height)`` at the ``(x, y)`` positions.

        Where ``x``, ``y``, ``width`` and ``height`` arguments are NumPy arrays, lists or scalar values.
        If ``height`` is None, it is set to the same value as width.
        """
        args = []
        buffers = []

        populate_args(x, args, buffers)
        populate_args(y, args, buffers)
        populate_args(width, args, buffers)

        if height is None:
            args.append(args[-1])
        else:
            populate_args(height, args, buffers)

        self._canvas_manager.send_draw_command(
            self, COMMANDS["fillRects"], args, buffers
        )

    def stroke_rects(self, x, y, width, height=None):
        """Draw a rectangular outlines of sizes ``(width, height)`` at the ``(x, y)`` positions.

        Where ``x``, ``y``, ``width`` and ``height`` arguments are NumPy arrays, lists or scalar values.
        If ``height`` is None, it is set to the same value as width.
        """
        args = []
        buffers = []

        populate_args(x, args, buffers)
        populate_args(y, args, buffers)
        populate_args(width, args, buffers)

        if height is None:
            args.append(args[-1])
        else:
            populate_args(height, args, buffers)

        self._canvas_manager.send_draw_command(
            self, COMMANDS["strokeRects"], args, buffers
        )

    def fill_styled_rects(self, x, y, width, height, color, alpha=1):
        """Draw filled and styled rectangles of sizes ``(width, height)`` at the ``(x, y)`` positions

        Where ``x``, ``y``, ``width`` and ``height`` arguments are NumPy arrays, lists or scalar values.
        If ``height`` is None, it is set to the same value as width.
        ``color`` is an (n_rect x 3) NumPy array with the colors and ``alpha`` is an (n_rect) NumPy array
        with the alpha channel values.
        """
        args = []
        buffers = []

        populate_args(x, args, buffers)
        populate_args(y, args, buffers)
        populate_args(width, args, buffers)

        if height is None:
            args.append(args[-1])
        else:
            populate_args(height, args, buffers)

        populate_args(color, args, buffers)
        populate_args(alpha, args, buffers)
        self._canvas_manager.send_draw_command(
            self, COMMANDS["fillStyledRects"], args, buffers
        )

    def stroke_styled_rects(self, x, y, width, height, color, alpha=1):
        """Draw rectangular styled outlines of sizes ``(width, height)`` at the ``(x, y)`` positions.of sizes ``(width, height)``

        Where ``x``, ``y``, ``width`` and ``height`` arguments are NumPy arrays, lists or scalar values.
        If ``height`` is None, it is set to the same value as width.
        ``color`` is an (n_rect x 3) NumPy array with the colors and ``alpha`` is an (n_rect) NumPy array
        with the alpha channel values.
        """
        args = []
        buffers = []

        populate_args(x, args, buffers)
        populate_args(y, args, buffers)
        populate_args(width, args, buffers)

        if height is None:
            args.append(args[-1])
        else:
            populate_args(height, args, buffers)

        populate_args(color, args, buffers)
        populate_args(alpha, args, buffers)
        self._canvas_manager.send_draw_command(
            self, COMMANDS["strokeStyledRects"], args, buffers
        )

    def clear_rect(self, x, y, width, height=None):
        """Clear the specified rectangular area of size ``(width, height)`` at the ``(x, y)`` position, making it fully transparent."""
        if height is None:
            height = width

        self._canvas_manager.send_draw_command(
            self, COMMANDS["clearRect"], [x, y, width, height]
        )

    # Arc methods
    def fill_arc(self, x, y, radius, start_angle, end_angle, anticlockwise=False):
        """Draw a filled arc centered at ``(x, y)`` with a radius of ``radius`` from ``start_angle`` to ``end_angle``."""
        self._canvas_manager.send_draw_command(
            self,
            COMMANDS["fillArc"],
            [x, y, radius, start_angle, end_angle, anticlockwise],
        )

    def fill_circle(self, x, y, radius):
        """Draw a filled circle centered at ``(x, y)`` with a radius of ``radius``."""
        self._canvas_manager.send_draw_command(
            self, COMMANDS["fillCircle"], [x, y, radius]
        )

    def stroke_arc(self, x, y, radius, start_angle, end_angle, anticlockwise=False):
        """Draw an arc outline centered at ``(x, y)`` with a radius of ``radius``."""
        self._canvas_manager.send_draw_command(
            self,
            COMMANDS["strokeArc"],
            [x, y, radius, start_angle, end_angle, anticlockwise],
        )

    def stroke_circle(self, x, y, radius):
        """Draw a circle centered at ``(x, y)`` with a radius of ``radius``."""
        self._canvas_manager.send_draw_command(
            self, COMMANDS["strokeCircle"], [x, y, radius]
        )

    def fill_arcs(self, x, y, radius, start_angle, end_angle, anticlockwise=False):
        """Draw filled arcs centered at ``(x, y)`` with a radius of ``radius``.

        Where ``x``, ``y``, ``radius`` and other arguments are NumPy arrays, lists or scalar values.
        """
        args = []
        buffers = []

        populate_args(x, args, buffers)
        populate_args(y, args, buffers)
        populate_args(radius, args, buffers)
        populate_args(start_angle, args, buffers)
        populate_args(end_angle, args, buffers)
        args.append(anticlockwise)

        self._canvas_manager.send_draw_command(
            self, COMMANDS["fillArcs"], args, buffers
        )

    def stroke_arcs(self, x, y, radius, start_angle, end_angle, anticlockwise=False):
        """Draw an arc outlines centered at ``(x, y)`` with a radius of ``radius``.

        Where ``x``, ``y``, ``radius`` and other arguments are NumPy arrays, lists or scalar values.
        """
        args = []
        buffers = []

        populate_args(x, args, buffers)
        populate_args(y, args, buffers)
        populate_args(radius, args, buffers)
        populate_args(start_angle, args, buffers)
        populate_args(end_angle, args, buffers)
        args.append(anticlockwise)

        self._canvas_manager.send_draw_command(
            self, COMMANDS["strokeArcs"], args, buffers
        )

    def fill_circles(self, x, y, radius):
        """Draw filled circles centered at ``(x, y)`` with a radius of ``radius``.

        Where ``x``, ``y``, ``radius`` and other arguments are NumPy arrays, lists or scalar values.
        """
        args = []
        buffers = []

        populate_args(x, args, buffers)
        populate_args(y, args, buffers)
        populate_args(radius, args, buffers)

        self._canvas_manager.send_draw_command(
            self, COMMANDS["fillCircles"], args, buffers
        )

    def stroke_circles(self, x, y, radius):
        """Draw a circle outlines centered at ``(x, y)`` with a radius of ``radius``.

        Where ``x``, ``y``, ``radius`` and other arguments are NumPy arrays, lists or scalar values.
        """
        args = []
        buffers = []

        populate_args(x, args, buffers)
        populate_args(y, args, buffers)
        populate_args(radius, args, buffers)

        self._canvas_manager.send_draw_command(
            self, COMMANDS["strokeCircles"], args, buffers
        )

    def fill_styled_circles(self, x, y, radius, color, alpha=1):
        """Draw a filled circles centered at ``(x, y)`` with a radius of ``radius``.

        Where ``x``, ``y``, ``radius``  and ``alpha`  are NumPy arrays, lists or scalar values.
        ``color`` must be an (nx3) NumPy arrays with the color for each circle.
        """
        args = []
        buffers = []

        populate_args(x, args, buffers)
        populate_args(y, args, buffers)
        populate_args(radius, args, buffers)
        populate_args(color, args, buffers)
        populate_args(alpha, args, buffers)
        self._canvas_manager.send_draw_command(
            self, COMMANDS["fillStyledCircles"], args, buffers
        )

    def stroke_styled_circles(self, x, y, radius, color, alpha=1):
        """Draw filled circles centered at ``(x, y)`` with a radius of ``radius``.

        Where ``x``, ``y``, ``radius``  and ``alpha``  are NumPy arrays, lists or scalar values.
        ``color`` must be an nx3 NumPy arrays with the color for each circle
        """
        args = []
        buffers = []

        populate_args(x, args, buffers)
        populate_args(y, args, buffers)
        populate_args(radius, args, buffers)
        populate_args(color, args, buffers)
        populate_args(alpha, args, buffers)

        self._canvas_manager.send_draw_command(
            self, COMMANDS["strokeStyledCircles"], args, buffers
        )

    def fill_styled_arcs(
        self, x, y, radius, start_angle, end_angle, color, alpha=1, anticlockwise=False
    ):
        """Draw filled and styled arcs centered at ``(x, y)`` with a radius of ``radius``.

        Where ``x``, ``y``, ``radius`` and other arguments are NumPy arrays, lists or scalar values.
        """
        args = []
        buffers = []

        populate_args(x, args, buffers)
        populate_args(y, args, buffers)
        populate_args(radius, args, buffers)
        populate_args(start_angle, args, buffers)
        populate_args(end_angle, args, buffers)
        args.append(anticlockwise)
        populate_args(color, args, buffers)
        populate_args(alpha, args, buffers)

        self._canvas_manager.send_draw_command(
            self, COMMANDS["fillStyledArcs"], args, buffers
        )

    def stroke_styled_arcs(
        self, x, y, radius, start_angle, end_angle, color, alpha=1, anticlockwise=False
    ):
        """Draw an styled arc outlines centered at ``(x, y)`` with a radius of ``radius``.

        Where ``x``, ``y``, ``radius`` and other arguments are NumPy arrays, lists or scalar values.
        """
        args = []
        buffers = []

        populate_args(x, args, buffers)
        populate_args(y, args, buffers)
        populate_args(radius, args, buffers)
        populate_args(start_angle, args, buffers)
        populate_args(end_angle, args, buffers)
        args.append(anticlockwise)
        populate_args(color, args, buffers)
        populate_args(alpha, args, buffers)

        self._canvas_manager.send_draw_command(
            self, COMMANDS["strokeStyledArcs"], args, buffers
        )

    # Polygon methods
    def fill_polygon(self, points):
        """Fill a polygon from a list of points ``[(x1, y1), (x2, y2), ..., (xn, yn)]``."""
        args = []
        buffers = []

        populate_args(points, args, buffers)

        self._canvas_manager.send_draw_command(
            self, COMMANDS["fillPolygon"], args, buffers
        )

    def stroke_polygon(self, points):
        """Draw polygon outline from a list of points ``[(x1, y1), (x2, y2), ..., (xn, yn)]``."""
        args = []
        buffers = []

        populate_args(points, args, buffers)

        self._canvas_manager.send_draw_command(
            self, COMMANDS["strokePolygon"], args, buffers
        )

    def fill_polygons(self, points, points_per_polygon=None):
        """ " Draw many filled polygons at once:

        Args:
            points (list or ndarray): The polygons points:

                The points can be specified as list or as ndarray:
                If the points are a list it must a an be a list of ndarrays,
                where each ndarray is a nx2 array of coordinates
                (n can be different for each entry)
                If the points are given as ndarray it must be either:

                - a 3d array: the shape of the array is (n_polyons, n_points_per_polygon, 2)
                - a 2d array: the shape of the array is (n,  2) and in additional  `points_per_polygon`
                  must be specified st. we know the the number of points for each individual polygon.
                  Note that the number of points in ``points`` must match the points_per_polygon.
                  array: ie: `np.sum(points_per_polygon) == points.shape[0]`

            points_per_polygon (ndarray):
                ndarray with number of points for each polygon. Must **only** be given if points are
                given as `flat` 2D array.

        """
        self._draw_polygons_or_linesegments(
            "fillPolygons", points, None, None, points_per_polygon, False, 3, "polygon"
        )

    def stroke_polygons(self, points, points_per_polygon=None):
        """ " Draw many stroked polygons at once:

        Args:
            points (list or ndarray): The polygons points:

                The points can be specified as list or as ndarray:
                If the points are a list it must a an be a list of ndarrays,
                where each ndarray is a nx2 array of coordinates
                (n can be different for each entry)
                If the points are given as ndarray it must be either:

                - a 3d array: the shape of the array is (n_polyons, n_points_per_polygon, 2)
                - a 2d array: the shape of the array is (n,  2) and in additional  `points_per_polygon`
                  must be specified st. we know the the number of points for each individual polygon.
                  Note that the number of points in ``points`` must match the points_per_polygon.
                  array: ie: `np.sum(points_per_polygon) == points.shape[0]`

            points_per_polygon (ndarray):
                ndarray with number of points for each polygon. Must **only** be given if points are
                given as `flat` 2D array.
        """
        self._draw_polygons_or_linesegments(
            "strokePolygons",
            points,
            None,
            None,
            points_per_polygon,
            False,
            3,
            "polygon",
        )

    def fill_styled_polygons(self, points, color, alpha=1, points_per_polygon=None):
        """ " Draw many filled polygons at once:

        Args:
            points (list or ndarray): The polygons points:

                The points can be specified as list or as ndarray:
                If the points are a list it must a an be a list of ndarrays,
                where each ndarray is a nx2 array of coordinates
                (n can be different for each entry)
                If the points are given as ndarray it must be either:

                - a 3d array: the shape of the array is (n_polyons, n_points_per_polygon, 2)
                - a 2d array: the shape of the array is (n,  2) and in additional  `points_per_polygon`
                  must be specified st. we know the the number of points for each individual polygon.
                  Note that the number of points in ``points`` must match the points_per_polygon.
                  array: ie: `np.sum(points_per_polygon) == points.shape[0]`

            color (ndarray)
                An (n_polyons,3) array with the color for each polygon
            alpha (ndarray,list,scalar):
                An  array with the alpha value for each polygon. Can be a scalar and the
                same value is used for all polygons
            points_per_polygon (ndarray):
                ndarray with number of points for each polygon. Must **only** be given if points are
                given as `flat` 2D array.

        """
        self._draw_polygons_or_linesegments(
            "fillStyledPolygons",
            points,
            color,
            alpha,
            points_per_polygon,
            True,
            3,
            "polygon",
        )

    def stroke_styled_polygons(self, points, color, alpha=1, points_per_polygon=None):
        """ " Draw many stroked polygons at once:

        Args:
            points (list or ndarray): The polygons points:

                The points can be specified as list or as ndarray:
                If the points are a list it must a an be a list of ndarrays,
                where each ndarray is a nx2 array of coordinates
                (n can be different for each entry)
                If the points are given as ndarray it must be either:

                - a 3d array: the shape of the array is (n_polyons, n_points_per_polygon, 2)
                - a 2d array: the shape of the array is (n,  2) and in additional  `points_per_polygon`
                  must be specified st. we know the the number of points for each individual polygon.
                  Note that the number of points in ``points`` must match the points_per_polygon.
                  array: ie: `np.sum(points_per_polygon) == points.shape[0]`

            color (ndarray)
                An (n_polyons,3) array with the color for each polygon
            alpha (ndarray,list,scalar):
                An  array with the alpha value for each polygon. Can be a scalar and the
                same value is used for all polygons
            points_per_polygon (ndarray):
                ndarray with number of points for each polygon. Must **only** be given if points are
                given as `flat` 2D array.

        """
        self._draw_polygons_or_linesegments(
            "strokeStyledPolygons",
            points,
            color,
            alpha,
            points_per_polygon,
            True,
            3,
            "polygon",
        )

    # Lines methods
    def stroke_line(self, x1, y1, x2, y2):
        """Draw a line from ``(x1, y1)`` to ``(x2, y2)``."""
        self._canvas_manager.send_draw_command(
            self, COMMANDS["strokeLine"], [x1, y1, x2, y2]
        )

    def stroke_lines(self, points):
        """Draw a path of consecutive lines from a list of points ``[(x1, y1), (x2, y2), ..., (xn, yn)]``."""
        args = []
        buffers = []

        populate_args(points, args, buffers)

        self._canvas_manager.send_draw_command(
            self, COMMANDS["strokeLines"], args, buffers
        )

    def stroke_styled_line_segments(
        self, points, color, alpha=1, points_per_line_segment=None
    ):
        """Draw many line segments at once:

        Args:
            points (list or ndarray): The line_segments points:

                The points can be specified as list or as ndarray:
                If the points are a list it must a an be a list of ndarrays,
                where each ndarray is a nx2 array of coordinates
                (n can be different for each entry)
                If the points are given as ndarray it must be either:

                - a 3d array: the shape of the array is (n_line_segments, n_points_per_polygon, 2)
                - a 2d array: the shape of the array is (n,  2) and in additional  `points_per_line_segment`
                    must be specified st. we know the the number of points for each individual line_segment.
                    Note that the number of points in ``points`` must match the points_per_line_segment.
                    array: ie: `np.sum(points_per_line_segment) == points.shape[0]`

            color (ndarray)
                An (n_line_segments,3) array with the color for each line_segment
            alpha (ndarray,list,scalar):
                An  array with the alpha value for each line_segment. Can be a scalar and the
                same value is used for all line_segments
            points_per_line_segment (ndarray):
                ndarray with number of points for each line_segment. Must **only** be given if points are
                given as `flat` 2D array.

        """
        self._draw_polygons_or_linesegments(
            "strokeStyledLineSegments",
            points,
            color,
            alpha,
            points_per_line_segment,
            True,
            2,
            "line_segment",
        )

    def stroke_line_segments(self, points, points_per_line_segment=None):
        """ Draw many stroked line_segments at once:

            Args:
                points (list or ndarray): The line_segments points:

                    The points can be specified as list or as ndarray:
                    If the points are a list it must a an be a list of ndarrays,
                    where each ndarray is a nx2 array of coordinates
                    (n can be different for each entry)
                    If the points are given as ndarray it must be either:

                    - a 3d array: the shape of the array is (n_line_segments, n_points_per_polygon, 2)
                    - a 2d array: the shape of the array is (n,  2) and in additional  `points_per_line_segment `
                      must be specified st. we know the the number of points for each individual polygon.
                      Note that the number of points in ``points`` must match the points_per_line_segment .
                      array: ie: `np.sum(points_per_line_segment  ) == points.shape[0]`\

                points_per_line_segment  (ndarray):
                    ndarray with number of points for each polygon. Must **only** be given if points are
                    given as `flat` 2D array.
        """
        self._draw_polygons_or_linesegments(
            "strokeLineSegments",
            points,
            None,
            None,
            points_per_line_segment,
            False,
            2,
            "line_segment",
        )

    # Paths methods
    def begin_path(self):
        """Call this method when you want to create a new path."""
        self._canvas_manager.send_draw_command(self, COMMANDS["beginPath"])

    def close_path(self):
        """Add a straight line from the current point to the start of the current path.

        If the shape has already been closed or has only one point, this function does nothing.
        This method doesn't draw anything to the canvas directly. You can render the path using the stroke() or fill() methods.
        """
        self._canvas_manager.send_draw_command(self, COMMANDS["closePath"])

    # https://developer.mozilla.org/en-US/docs/Web/API/CanvasRenderingContext2D/stroke
    # stroke(), stroke(path)
    def stroke(self, path2d: Path2D = None):
        """Stroke (outlines) the current path with the current ``stroke_style``.
        If @path2d is passed, that Path2D object will be rendered with the
        current ``stroke_style``"""

        if isinstance(path2d, Path2D):
            self._canvas_manager.send_draw_command(
                self,
                COMMANDS["strokePath"],
                [widget_serialization["to_json"](path2d, None)],
            )
        else:
            self._canvas_manager.send_draw_command(self, COMMANDS["stroke"])

    def fill(self, rule_or_path="nonzero"):
        """Fill the current path with the current ``fill_style`` and given the rule, or fill the given Path2D.

        Possible rules are ``nonzero`` and ``evenodd``.
        """
        if isinstance(rule_or_path, Path2D):
            self._canvas_manager.send_draw_command(
                self,
                COMMANDS["fillPath"],
                [widget_serialization["to_json"](rule_or_path, None)],
            )
        else:
            self._canvas_manager.send_draw_command(
                self, COMMANDS["fill"], [rule_or_path]
            )

    def move_to(self, x, y):
        """Move the "pen" to the given ``(x, y)`` coordinates."""
        self._canvas_manager.send_draw_command(self, COMMANDS["moveTo"], [x, y])

    def line_to(self, x, y):
        """Add a straight line to the current path by connecting the path's last point to the specified ``(x, y)`` coordinates.

        Like other methods that modify the current path, this method does not directly render anything. To
        draw the path onto the canvas, you can use the fill() or stroke() methods.
        """
        self._canvas_manager.send_draw_command(self, COMMANDS["lineTo"], [x, y])

    def rect(self, x, y, width, height):
        """Add a rectangle of size ``(width, height)`` at the ``(x, y)`` position in the current path."""
        self._canvas_manager.send_draw_command(
            self, COMMANDS["rect"], [x, y, width, height]
        )

    def arc(self, x, y, radius, start_angle, end_angle, anticlockwise=False):
        """Add a circular arc centered at ``(x, y)`` with a radius of ``radius`` to the current path.

        The path starts at ``start_angle`` and ends at ``end_angle``, and travels in the direction given by
        ``anticlockwise`` (defaulting to clockwise: ``False``).
        """
        self._canvas_manager.send_draw_command(
            self, COMMANDS["arc"], [x, y, radius, start_angle, end_angle, anticlockwise]
        )

    def ellipse(
        self,
        x,
        y,
        radius_x,
        radius_y,
        rotation,
        start_angle,
        end_angle,
        anticlockwise=False,
    ):
        """Add an ellipse centered at ``(x, y)`` with the radii ``radius_x`` and ``radius_y`` to the current path.

        The path starts at ``start_angle`` and ends at ``end_angle``, and travels in the direction given by
        ``anticlockwise`` (defaulting to clockwise: ``False``).
        """
        self._canvas_manager.send_draw_command(
            self,
            COMMANDS["ellipse"],
            [x, y, radius_x, radius_y, rotation, start_angle, end_angle, anticlockwise],
        )

    def arc_to(self, x1, y1, x2, y2, radius):
        """Add a circular arc to the current path.

        Using the given control points ``(x1, y1)`` and ``(x2, y2)`` and the ``radius``.
        """
        self._canvas_manager.send_draw_command(
            self, COMMANDS["arcTo"], [x1, y1, x2, y2, radius]
        )

    def quadratic_curve_to(self, cp1x, cp1y, x, y):
        """Add a quadratic Bezier curve to the current path.

        It requires two points: the first one is a control point and the second one is the end point.
        The starting point is the latest point in the current path, which can be changed using move_to()
        before creating the quadratic Bezier curve.
        """
        self._canvas_manager.send_draw_command(
            self, COMMANDS["quadraticCurveTo"], [cp1x, cp1y, x, y]
        )

    def bezier_curve_to(self, cp1x, cp1y, cp2x, cp2y, x, y):
        """Add a cubic Bezier curve to the current path.

        It requires three points: the first two are control points and the third one is the end point.
        The starting point is the latest point in the current path, which can be changed using move_to()
        before creating the Bezier curve.
        """
        self._canvas_manager.send_draw_command(
            self, COMMANDS["bezierCurveTo"], [cp1x, cp1y, cp2x, cp2y, x, y]
        )

    # Text methods
    def fill_text(self, text, x, y, max_width=None):
        """Fill a given text at the given ``(x, y)`` position. Optionally with a maximum width to draw."""
        self._canvas_manager.send_draw_command(
            self, COMMANDS["fillText"], [text, x, y, max_width]
        )

    def stroke_text(self, text, x, y, max_width=None):
        """Stroke a given text at the given ``(x, y)`` position. Optionally with a maximum width to draw."""
        self._canvas_manager.send_draw_command(
            self, COMMANDS["strokeText"], [text, x, y, max_width]
        )

    # Line methods
    def get_line_dash(self):
        """Return the current line dash pattern array containing an even number of non-negative numbers."""
        return self._line_dash

    def set_line_dash(self, segments):
        """Set the current line dash pattern."""
        if len(segments) % 2:
            self._line_dash = segments + segments
        else:
            self._line_dash = segments
        self._canvas_manager.send_draw_command(
            self, COMMANDS["setLineDash"], [self._line_dash]
        )

    # Image methods
    def draw_image(self, image, x=0, y=0, width=None, height=None):
        """Draw an ``image`` on the Canvas at the coordinates (``x``, ``y``) and scale it to (``width``, ``height``)."""
        if not isinstance(image, (Canvas, MultiCanvas, Image)):
            raise TypeError(
                "The image argument should be an Image, a Canvas or a MultiCanvas widget"
            )

        if width is not None and height is None:
            height = width

        serialized_image = widget_serialization["to_json"](image, None)

        self._canvas_manager.send_draw_command(
            self, COMMANDS["drawImage"], [serialized_image, x, y, width, height]
        )

    def put_image_data(self, image_data, x=0, y=0):
        """Draw an image on the Canvas.

        ``image_data`` should be  a NumPy array containing the image to draw and ``x`` and ``y`` the pixel position where to
        draw. Unlike the CanvasRenderingContext2D.putImageData method, this method **is** affected by the canvas transformation
        matrix, and supports transparency.
        """
        image_buffer = binary_image(image_data)
        self._canvas_manager.send_draw_command(
            self, COMMANDS["putImageData"], [x, y], [image_buffer]
        )

    def create_image_data(self, width, height):
        """Create a NumPy array of shape (width, height, 4) representing a table of pixel colors."""
        return np.zeros((width, height, 4), dtype=int)

    # Clipping
    def clip(self):
        """Turn the path currently being built into the current clipping path.

        You can use clip() instead of close_path() to close a path and turn it into a clipping
        path instead of stroking or filling the path.
        """
        self._canvas_manager.send_draw_command(self, COMMANDS["clip"])

    # Transformation methods
    def save(self):
        """Save the entire state of the canvas."""
        self._canvas_manager.send_draw_command(self, COMMANDS["save"])

    def restore(self):
        """Restore the most recently saved canvas state."""
        self._canvas_manager.send_draw_command(self, COMMANDS["restore"])

    def translate(self, x, y):
        """Move the canvas and its origin on the grid.

        ``x`` indicates the horizontal distance to move,
        and ``y`` indicates how far to move the grid vertically.
        """
        self._canvas_manager.send_draw_command(self, COMMANDS["translate"], [x, y])

    def rotate(self, angle):
        """Rotate the canvas clockwise around the current origin by the ``angle`` number of radians."""
        self._canvas_manager.send_draw_command(self, COMMANDS["rotate"], [angle])

    def scale(self, x, y=None):
        """Scale the canvas units by ``x`` horizontally and by ``y`` vertically. Both parameters are real numbers.

        If ``y`` is not provided, it is defaulted to the same value as ``x``.
        Values that are smaller than 1.0 reduce the unit size and values above 1.0 increase the unit size.
        Values of 1.0 leave the units the same size.
        """
        if y is None:
            y = x
        self._canvas_manager.send_draw_command(self, COMMANDS["scale"], [x, y])

    def transform(self, a, b, c, d, e, f):
        """Multiply the current transformation matrix with the matrix described by its arguments.

        The transformation matrix is described by:
        ``[[a, c, e], [b, d, f], [0, 0, 1]]``.
        """
        self._canvas_manager.send_draw_command(
            self, COMMANDS["transform"], [a, b, c, d, e, f]
        )

    def set_transform(self, a, b, c, d, e, f):
        """Reset the current transform to the identity matrix, and then invokes the transform() method with the same arguments.

        This basically undoes the current transformation, then sets the specified transform, all in one step.
        """
        self._canvas_manager.send_draw_command(
            self, COMMANDS["setTransform"], [a, b, c, d, e, f]
        )

    def reset_transform(self):
        """Reset the current transform to the identity matrix.

        This is the same as calling: set_transform(1, 0, 0, 1, 0, 0).
        """
        self._canvas_manager.send_draw_command(self, COMMANDS["resetTransform"])

    # Extras
    def clear(self):
        """Clear the entire canvas. This is the same as calling ``clear_rect(0, 0, canvas.width, canvas.height)``."""
        self._canvas_manager.send_draw_command(self, COMMANDS["clear"])

    def flush(self):
        """Flush all the cached commands and clear the cache."""
        self._canvas_manager.flush()

    # Events
    def on_client_ready(self, callback, remove=False):
        """Register a callback that will be called when a new client is ready to receive draw commands.

        When a new client connects to the kernel he will get an empty Canvas (because the canvas is
        almost stateless, the new client does not know what draw commands were previously sent). So
        this function is useful for replaying your drawing whenever a new client connects and is
        ready to receive draw commands.
        """
        self._client_ready_callbacks.register_callback(callback, remove=remove)

    def on_mouse_move(self, callback, remove=False):
        """Register a callback that will be called on mouse move."""
        self._mouse_move_callbacks.register_callback(callback, remove=remove)

    def on_mouse_down(self, callback, remove=False):
        """Register a callback that will be called on mouse click down."""
        self._mouse_down_callbacks.register_callback(callback, remove=remove)

    def on_mouse_up(self, callback, remove=False):
        """Register a callback that will be called on mouse click up."""
        self._mouse_up_callbacks.register_callback(callback, remove=remove)

    def on_mouse_out(self, callback, remove=False):
        """Register a callback that will be called on mouse out of the canvas."""
        self._mouse_out_callbacks.register_callback(callback, remove=remove)

    def on_mouse_wheel(self, callback, remove=False):
        """Register a callback that will be called on mouse wheel movement."""
        self._mouse_wheel_callbacks.register_callback(callback, remove=remove)

    def on_touch_start(self, callback, remove=False):
        """Register a callback that will be called on touch start (new finger on the screen)."""
        self._touch_start_callbacks.register_callback(callback, remove=remove)

    def on_touch_end(self, callback, remove=False):
        """Register a callback that will be called on touch end (a finger is not touching the screen anymore)."""
        self._touch_end_callbacks.register_callback(callback, remove=remove)

    def on_touch_move(self, callback, remove=False):
        """Register a callback that will be called on touch move (finger moving on the screen)."""
        self._touch_move_callbacks.register_callback(callback, remove=remove)

    def on_touch_cancel(self, callback, remove=False):
        """Register a callback that will be called on touch cancel."""
        self._touch_cancel_callbacks.register_callback(callback, remove=remove)

    def on_key_down(self, callback, remove=False):
        """Register a callback that will be called on keyboard event."""
        self._key_down_callbacks.register_callback(callback, remove=remove)

    def __setattr__(self, name, value):
        super(Canvas, self).__setattr__(name, value)

        if name in self.ATTRS:
            # If it's a Widget we need to serialize it
            if isinstance(value, Widget):
                value = widget_serialization["to_json"](value, None)

            self._canvas_manager.send_command(
                self, [COMMANDS["set"], [self.ATTRS[name], value]]
            )

    def _handle_frontend_event(self, _, content, buffers):
        if content.get("event", "") == "client_ready":
            self._client_ready_callbacks()

        if content.get("event", "") == "mouse_move":
            self._mouse_move_callbacks(content["x"], content["y"])
        if content.get("event", "") == "mouse_down":
            self._mouse_down_callbacks(content["x"], content["y"])
        if content.get("event", "") == "mouse_up":
            self._mouse_up_callbacks(content["x"], content["y"])
        if content.get("event", "") == "mouse_out":
            self._mouse_out_callbacks(content["x"], content["y"])
        if content.get("event", "") == "mouse_wheel":
            self._mouse_wheel_callbacks(content["x"], content["y"])

        if content.get("event", "") == "touch_start":
            self._touch_start_callbacks(
                [(touch["x"], touch["y"]) for touch in content["touches"]]
            )
        if content.get("event", "") == "touch_end":
            self._touch_end_callbacks(
                [(touch["x"], touch["y"]) for touch in content["touches"]]
            )
        if content.get("event", "") == "touch_move":
            self._touch_move_callbacks(
                [(touch["x"], touch["y"]) for touch in content["touches"]]
            )
        if content.get("event", "") == "touch_cancel":
            self._touch_cancel_callbacks(
                [(touch["x"], touch["y"]) for touch in content["touches"]]
            )

        if content.get("event", "") == "key_down":
            self._key_down_callbacks(
                content["key"],
                content["shift_key"],
                content["ctrl_key"],
                content["meta_key"],
            )

    def _draw_polygons_or_linesegments(
        self,
        cmd,
        points,
        color,
        alpha,
        points_per_item,
        with_style,
        min_elements,
        item_name,
    ):
        args = []
        buffers = []

        (
            num_polygons,
            flat_points,
            points_per_item,
        ) = _serialize_list_of_polygons_or_linestrokes(
            points=points,
            points_per_item=points_per_item,
            item_name=item_name,
            min_elements=min_elements,
        )

        if with_style:
            color = np.require(color, requirements=["C"], dtype="uint8")
            if color.ndim != 1:
                color = color.ravel()

        populate_args(num_polygons, args, buffers)
        populate_args(flat_points, args, buffers)
        populate_args(points_per_item, args, buffers)
        if with_style:
            populate_args(color, args, buffers)
            populate_args(alpha, args, buffers)
        self._canvas_manager.send_draw_command(self, COMMANDS[cmd], args, buffers)


class RoughCanvas(Canvas):
    """Create a RoughCanvas widget. It gives a hand-drawn-like style to your drawings.

    Args:
        width (int): The width (in pixels) of the canvas
        height (int): The height (in pixels) of the canvas
    """

    _model_name = Unicode("RoughCanvasModel").tag(sync=True)
    _view_name = Unicode("CanvasView").tag(sync=True)

    #: (str) Sets the appearance of the filling, possible values are ``'hachure'``, ``'solid'``, ``'zigzag'``,
    #: ``'cross-hatch'``, ``'dots'``, ``'sunburst'``, ``'dashed'``, ``'zigzag-line'``.
    #: Default to ``'hachure'``.
    rough_fill_style = Enum(
        [
            "hachure",
            "solid",
            "zigzag",
            "cross-hatch",
            "dots",
            "sunburst",
            "dashed",
            "zigzag-line",
        ],
        default_value="hachure",
    )

    #: (float) Numerical value indicating how rough the drawing is. A rectangle with the roughness of 0 would be a perfect rectangle.
    #: There is no upper limit to this value, but a value over 10 is mostly useless.
    #: Default to ``'1'``.
    roughness = Float(1)

    #: (float) Numerical value indicating how curvy the lines are when drawing a sketch. A value of 0 will cause straight lines.
    #: Default to ``'1'``.
    bowing = Float(1)

    ROUGH_ATTRS = {
        "rough_fill_style": 100,
        "roughness": 101,
        "bowing": 102,
    }

    def __setattr__(self, name, value):
        super(RoughCanvas, self).__setattr__(name, value)

        if name in self.ROUGH_ATTRS:
            self._canvas_manager.send_command(
                self, [COMMANDS["set"], [self.ROUGH_ATTRS[name], value]]
            )


class MultiCanvas(_CanvasBase):
    """Create a MultiCanvas widget with n_canvases Canvas widgets.

    Args:
        n_canvases (int): The number of canvases to create
        width (int): The width (in pixels) of the canvases
        height (int): The height (in pixels) of the canvases
    """

    _model_name = Unicode("MultiCanvasModel").tag(sync=True)
    _view_name = Unicode("MultiCanvasView").tag(sync=True)

    _canvases = List(Instance(Canvas)).tag(sync=True, **widget_serialization)

    def __init__(self, n_canvases=3, *args, **kwargs):
        """Constructor."""
        super(MultiCanvas, self).__init__(
            *args, _canvases=[Canvas() for _ in range(n_canvases)], **kwargs
        )

        # The latest canvas receives events (interaction layer)
        self.on_msg(self._canvases[-1]._handle_frontend_event)

    def __getitem__(self, key):
        """Access one of the Canvas instances."""
        return self._canvases[key]

    def __setattr__(self, name, value):
        super(MultiCanvas, self).__setattr__(name, value)

        if name in ("width", "height"):
            for layer in self._canvases:
                setattr(layer, name, value)

        if name == "caching":
            self._canvas_manager._caching = value

            warnings.warn(
                "caching is deprecated and will be removed in a future release, please use hold_canvas() instead.",
                DeprecationWarning,
            )

    def __getattr__(self, name):
        if name in ("width", "height"):
            return getattr(self._canvases[0], name)

        if name == "caching":
            warnings.warn(
                "caching is deprecated and will be removed in a future release, please use hold_canvas() instead.",
                DeprecationWarning,
            )
            return self._canvas_manager._caching

        raise AttributeError(f"'MultiCanvas' object has no attribute '{name}'")

    def on_client_ready(self, callback, remove=False):
        """Register a callback that will be called when a new client is ready to receive draw commands.

        When a new client connects to the kernel he will get an empty Canvas (because the canvas is
        almost stateless, the new client does not know what draw commands were previously sent). So
        this function is useful for replaying your drawing whenever a new client connects and is
        ready to receive draw commands.
        """
        self._canvases[-1].on_client_ready(callback, remove=remove)

    def on_mouse_move(self, callback, remove=False):
        """Register a callback that will be called on mouse move."""
        self._canvases[-1].on_mouse_move(callback, remove=remove)

    def on_mouse_down(self, callback, remove=False):
        """Register a callback that will be called on mouse click down."""
        self._canvases[-1].on_mouse_down(callback, remove=remove)

    def on_mouse_up(self, callback, remove=False):
        """Register a callback that will be called on mouse click up."""
        self._canvases[-1].on_mouse_up(callback, remove=remove)

    def on_mouse_out(self, callback, remove=False):
        """Register a callback that will be called on mouse out of the canvas."""
        self._canvases[-1].on_mouse_out(callback, remove=remove)

    def on_mouse_wheel(self, callback, remove=False):
        """Register a callback that will be called on mouse wheel movement."""
        self._canvases[-1].on_mouse_wheel(callback, remove=remove)

    def on_touch_start(self, callback, remove=False):
        """Register a callback that will be called on touch start (new finger on the screen)."""
        self._canvases[-1].on_touch_start(callback, remove=remove)

    def on_touch_end(self, callback, remove=False):
        """Register a callback that will be called on touch end (a finger is not touching the screen anymore)."""
        self._canvases[-1].on_touch_end(callback, remove=remove)

    def on_touch_move(self, callback, remove=False):
        """Register a callback that will be called on touch move (finger moving on the screen)."""
        self._canvases[-1].on_touch_move(callback, remove=remove)

    def on_touch_cancel(self, callback, remove=False):
        """Register a callback that will be called on touch cancel."""
        self._canvases[-1].on_touch_cancel(callback, remove=remove)

    def on_key_down(self, callback, remove=False):
        """Register a callback that will be called on keyboard event."""
        self._canvases[-1].on_key_down(callback, remove=remove)

    def clear(self):
        """Clear the Canvas."""
        for layer in self._canvases:
            layer.clear()

    def flush(self):
        """Flush all the cached commands and clear the cache."""
        self._canvas_manager.flush()


class MultiRoughCanvas(MultiCanvas):
    """Create a MultiRoughCanvas widget with n_canvases RoughCanvas widgets.

    Args:
        n_canvases (int): The number of rough canvases to create
        width (int): The width (in pixels) of the canvases
        height (int): The height (in pixels) of the canvases
    """

    _canvases = List(Instance(RoughCanvas)).tag(sync=True, **widget_serialization)

    def __init__(self, n_canvases=3, *args, **kwargs):
        """Constructor."""
        super(MultiCanvas, self).__init__(
            *args, _canvases=[RoughCanvas() for _ in range(n_canvases)], **kwargs
        )


@contextmanager
def hold_canvas(canvas=None):
    """Hold any drawing, and perform all commands in a single shot at the end.

    This is way more efficient than sending commands one by one.
    """
    if canvas is not None:
        warnings.warn(
            "hold_canvas does not take a canvas as parameter anymore, please use hold_canvas() instead.",
            DeprecationWarning,
        )

    orig_caching = _CANVAS_MANAGER._caching

    _CANVAS_MANAGER._caching = True
    yield
    _CANVAS_MANAGER.flush()

    if not orig_caching:
        _CANVAS_MANAGER._caching = False



from .offscreen_canvas_core import OffscreenCanvasCore
from contextlib import contextmanager
from functools import partial,partialmethod
import pyjs
import numpy as np
from numbers import Number
from pathlib import Path
from IPython.display import display
from ipywidgets import Image as IpywidgetImage
import io
import PIL

@contextmanager
def hold_canvas(canvas):
    yield None

class OffscreenCanvas(OffscreenCanvasCore):
    


    def __init__(self, *args, **kwargs): 
        super().__init__(*args, **kwargs)

        # create buffers for the batch API:
        # All buffers are of type float32 since for batch calls we only use floats.
        # We preallocate buffers and create js views for them.
        # In each batch call we check if the buffer is large enough and resize it if necessary.
        # we then fill the buffers with the data, and fill the last buffer with the sizes of
        # the other buffers.
        # the number of buffers can be fixed since we only use a fixed number of arguments in the batch calls.
        initial_buffer_size = 10
        n_buffers = 8 # max number of args / buffers we need at the same time
        self._buffers = [ np.zeros(initial_buffer_size, dtype=np.float32) for _ in range(n_buffers)]
        self._js_buffers = [
            pyjs.buffer_to_js_typed_array(fbuffer, view=True) for fbuffer in self._buffers
        ]

    def initialize(self):
        """ After the canvas has been displayed, we need to call this method to initialize the canvas.

            After the canvas has been displayed, we need to call this method to initialize the canvas.
            This method can only be used when the display method has been called in a **different cell**.

            # Cell1:
            ```python
            from ipycanvas.compat import Canvas
            from IPython.display import display
            canvas = Canvas(width=800, height=600)
            display(canvas)
            ```

            # Cell2:
            ```python
            canvas.initialize()
            ```

            For more information, see the `async_initialize` method.

        """
        super().initialize()
        if  self._canvas is None:
            raise RuntimeError("Canvas is not displayed yet")
        self._ctx = self._canvas.getContext("2d")

    async def async_initialize(self):
        """
        
                If we want to use the canvas in the same cell where it 
                was created **and displayed** we need to call this async function.
                While this sounds a bit counterintuitive, it is necessary because the
                offscreen canvas is created as regular canvas in the main-thread,
                and transfered to and recieved by the worker-thread. This transfering mechanism (in particular the
                receiving part) would be blocked by the cell execution.
                To get a chance to  receive the canvas in the worker-thread (ie where
                the kernel is running), we need to do run some asyc code (with some sleeping in between)
                Note that for an ordinary canvas, `async_initialize` is a no-op.

            from ipycanvas.compat import Canvas
            from IPython.display import display
            canvas = Canvas(width=800, height=600)
            display(canvas)
            await canvas.async_initialize()

            # canvas is now ready to use
            canvas.fill_style = 'red'


        """
        
        await super().async_initialize()
        if  self._canvas is None:
            raise RuntimeError("Canvas is not displayed yet")
        self._ctx = self._canvas.getContext("2d")
    
    async def display(self):
        """ shorthand for displaying the canvas and then initializing it.
            See `async_initialize` for more information.
        """
        display(self)
        await self.async_initialize()

  
    # ipycanvas api
    def clear(self):
        self._ctx.clearRect(0, 0, self._canvas.width, self._canvas.height)
    
    # for compatibility with the regular canvas
    def sleep(self, seconds):
        """in the non-lite version this sleeps in the fronend / canvas, but not in the kernel.
        THis make little sense  in the offscreen canvas version, since the canvas **is** the frontend.
        """
        pass

    def create_linear_gradient(self, x0, y0, x1, y1, color_stops):
        gradient = self._ctx.createLinearGradient(x0, y0, x1, y1)
        for offset, color in color_stops:
            gradient.addColorStop(offset, color)
        return gradient
    
    def create_radial_gradient(self, x0, y0, r0, x1, y1, r1, color_stops):
        """Create a radial gradient."""
        gradient = self._ctx.createRadialGradient(x0, y0, r0, x1, y1, r1)
        for offset, color in color_stops:
            gradient.addColorStop(offset, color)
        return gradient

    def create_pattern(self, image, repetition='repeat'):
        if isinstance(image, OffscreenCanvasCore):
            # if the image is an OffscreenCanvasCore, we need to convert it to a js image
            image = image._canvas
        else:
            raise NotImplementedError("create_pattern only supports OffscreenCanvas images at the moment")
        
        pattern = self._ctx.createPattern(image, repetition)
        return pattern

    def fill_rect(self, x, y, width, height):
        self._ctx.fillRect(x, y, width, height)
    def stroke_rect(self, x, y, width, height):
        self._ctx.strokeRect(x, y, width, height)
    def clear_rect(self):
        self._ctx.clearRect(x, y, width, height)
    def fill_arc(self, x, y, radius, start_angle, end_angle, anticlockwise=False):
        self._ctx.fillArc(x, y, radius, start_angle, end_angle, anticlockwise)
    def fill_circle(self, x, y, radius):
        self._ctx.fillCircle(x, y, radius)
    def stroke_arc(self, x, y, radius, start_angle, end_angle, anticlockwise=False):
        self._ctx.strokeArc(x, y, radius, start_angle, end_angle, anticlockwise)
    def stroke_circle(self, x, y, radius):
        self._ctx.strokeCircle(x, y, radius)

    def fill_polygon(self, points):
        n_points = self._fill_buffer_with_points(0, points)
        self._ctx.fillPolygon(n_points, self._js_buffers[0] )

    def stroke_polygon(self, points):
        n_points = self._fill_buffer_with_points(0, points)
        self._ctx.strokePolygon(n_points ,self._js_buffers[0])

    def fill_and_stroke_polygon(self, points):
        n_points = self._fill_buffer_with_points(0, points)
        self._ctx.fillAndStrokePolygon(n_points, self._js_buffers[0])

    def stroke_line(self, x1, y1, x2, y2):
        self._ctx.strokeLine(x1, y1, x2, y2)
    def begin_path(self):
        self._ctx.beginPath()
    def close_path(self):
        self._ctx.closePath()
    def stroke(self):
        self._ctx.stroke()
    def fill(self):
        self._ctx.fill()
    def move_to(self, x, y):
        self._ctx.moveTo(x, y)
    def line_to(self, x, y):
        self._ctx.lineTo(x, y)
    def rect(self, x, y, width, height):
        self._ctx.rect(x, y, width, height)
    def arc(self, x, y, radius, start_angle, end_angle, anticlockwise=False):
        self._ctx.arc(x, y, radius, start_angle, end_angle, anticlockwise)  
    def ellipse(self, x, y, radius_x, radius_y, rotation=0, start_angle=0, end_angle=2 * 3.14159, anticlockwise=False):
        self._ctx.ellipse(x, y, radius_x, radius_y, rotation, start_angle, end_angle, anticlockwise)
    def arc_to(self, x1, y1, x2, y2, radius):
        self._ctx.arcTo(x1, y1, x2, y2, radius)
    def quadratic_curve_to(self, cp_x, cp_y, to_x, to_y):
        self._ctx.quadraticCurveTo(cp_x, cp_y, to_x, to_y)
    def bezier_curve_to(self, cp1_x, cp1_y, cp2_x, cp2_y, to_x, to_y):
        self._ctx.bezierCurveTo(cp1_x, cp1_y, cp2_x, cp2_y, to_x, to_y)
    def fill_text(self, text, x, y, max_width=None):
        if max_width is not None:
            self._ctx.fillText(text, x, y, max_width)
        else:
            self._ctx.fillText(text, x, y)
    def stroke_text(self, text, x, y, max_width=None):
        if max_width is not None:
            self._ctx.strokeText(text, x, y, max_width)
        else:
            self._ctx.strokeText(text, x, y)
    def get_line_dash(self):
        raise NotImplementedError("get_line_dash is not implemented in the offscreen canvas version yet")
    def set_line_dash(self, segments):
        raise NotImplementedError("set_line_dash is not implemented in the offscreen canvas version yet")
    def draw_image(self):
        raise NotImplementedError("draw_image is not implemented in the offscreen canvas version yet")
    def put_image_data(self, image_data, dx, dy, dirty_x=None, dirty_y=None, dirty_width=None, dirty_height=None):
        raise NotImplementedError("put_image_data is not implemented in the offscreen canvas version yet")
    def create_image_data(self, sw=None, sh=None):
        raise NotImplementedError("create_image_data is not implemented in the offscreen canvas version yet")
    def clip(self):
        self._ctx.clip()
    def save(self):
        self._ctx.save()
    def restore(self):
        self._ctx.restore()
    def translate(self, x, y):
        self._ctx.translate(x, y)
        
    def rotate(self, angle):
        self._ctx.rotate(angle)
        
    def scale(self, x, y):
        self._ctx.scale(x, y)
        
    def transform(self, a, b, c, d, e, f):
        self._ctx.transform(a, b, c, d, e, f)
        
    def set_transform(self, a, b, c, d, e, f):
        self._ctx.setTransform(a, b, c, d, e, f)
    def reset_transform(self):
        self._ctx.resetTransform()  
    def clear(self):
        """Clear the canvas."""
        self._ctx.clearRect(0, 0, self._canvas.width, self._canvas.height)
        
    def flush(self):
        """Flush the canvas. In the offscreen canvas version this does nothing."""
        pass
    
    def draw_image(self, image, dx, dy, dw=None, dh=None):
        """Draw an image on the canvas."""
        if isinstance(image, OffscreenCanvasCore):
            # if the image is an OffscreenCanvasCore, we need to convert it to a js image
            drawable_image = image._canvas

        elif isinstance(image, IpywidgetImage):
            if dw is not None and dh is not None:
                raise NotImplementedError("ipywidget.Image does not support width and height parameters in draw_image")

            # convert to stream an open with pillow
            data_stream = io.BytesIO(image.value)
            pil_img = PIL.Image.open(data_stream)

            # convert to RGBA if not already in that mode
            if pil_img.mode != 'RGBA':
                pil_img = pil_img.convert('RGBA'
            )

            # convert to numpy
            img_rgba = np.array(pil_img) 

            # create a js array view (this will be of the type Uint8)
            js_arr = pyjs.buffer_to_js_typed_array(img_rgba.ravel(), view=True)
            # convert to Uint8ClampedArray without copying the data
            js_arr = pyjs.js.Uint8ClampedArray.new(js_arr.buffer, js_arr.byteOffset, js_arr.length)

            # create settings to ensure the pixel format is correct
            settings = pyjs.js_object()
            settings.pixelFormat = "rgba-unorm8"

            # create the ImageData object
            image = pyjs.js.ImageData.new(js_arr, pil_img.width, pil_img.height, settings)

            # create an OffscreenCanvas and draw the image data on it
            offscreen_canvas = pyjs.js.OffscreenCanvas.new( pil_img.width, pil_img.height)
            ctx = offscreen_canvas.getContext('2d')

            # put the image data on the offscreen canvas
            ctx.putImageData(image,0,0)
            drawable_image = offscreen_canvas
          
        else:
            raise NotImplementedError("draw_image only supports OffscreenCanvas and ipywidget.Image as images at the moment")

        if dw is not None and dh is not None:
            self._ctx.drawImage(drawable_image, dx, dy, dw, dh)
        else:
            self._ctx.drawImage(drawable_image, dx, dy)    
    
    def put_image_data(self, image_data, x=0,y=0):
        n_dim = image_data.ndim
        width = image_data.shape[0]
        height = image_data.shape[1]
        if n_dim == 2:
            # grayscale image
            # convert to RGBA
            image_data = np.stack((image_data,)*3, axis=-1)
            image_data = np.concatenate((image_data, np.full(image_data.shape[:2] + (1,), 255)), axis=-1)

            
        elif n_dim == 3 and image_data.shape[-1] == 1:
            # single channel image (grayscale)
            # convert to RGBA
            image_data = np.concatenate((image_data, np.full(image_data.shape[:2] + (1,), 255)), axis=-1)

        elif n_dim == 3: 
            # RGB image
            if image_data.shape[-1] == 3:
                # add alpha channel
                image_data = np.concatenate((image_data, np.full(image_data.shape[:2] + (1,), 255)), axis=-1)
            elif image_data.shape[-1] == 4:
                # already RGBA
                pass
            else:
                raise ValueError("Image data must be 2D or 3D with 3 or 4 channels")

        image_data = np.require(image_data, requirements='C', dtype=np.uint8)

        # create a js array view (this will be of the type Uint8)
        js_arr = pyjs.buffer_to_js_typed_array(image_data.ravel(), view=True)
        # convert to Uint8ClampedArray without copying the data
        js_arr = pyjs.js.Uint8ClampedArray.new(js_arr.buffer, js_arr.byteOffset, js_arr.length)

        # create settings to ensure the pixel format is correct
        settings = pyjs.js_object()
        settings.pixelFormat = "rgba-unorm8"

        image = pyjs.js.ImageData.new(js_arr, width, height, settings)

        # put_image_data on the offscreen canvas
        self._ctx.putImageData(image, x, y)



    # ensure that the buffer at index hast at least the given size.
    def _ensure_size(self, index, size):
        buffer_size = len(self._buffers[index])
        if size > buffer_size:
            # resize the buffer
            new_size = max(size, buffer_size * 2)
            new_buffer = np.zeros(new_size, dtype=np.float32)
            self._buffers[index] =  new_buffer
            self._js_buffers[index] = pyjs.buffer_to_js_typed_array(new_buffer, view=True)

    # helper to add a list / array of points to a buffer 
    # (for instance when drawing a polygon from a list of points)
    def _fill_buffer_with_points(self,index, points):
        points = np.require(points, requirements='C', dtype=np.float32)
        n_points = int(points.shape[0])
        n_points2 = 2 * n_points
        self._ensure_size(index, n_points2)
        self._buffers[index][:n_points2] = points.flatten()
        return n_points

    # helper to fill a buffer with a scalar value
    def _fill_buffer_with_scalars(self, index, value):
        # is number ? 
        if isinstance(value, Number):
            self._ensure_size(index, 1)
            self._buffers[index][0] = value
            return 1
        else: # assume iterable
            value = np.require(value, requirements='C', dtype=np.float32)
            n_values = int(value.shape[0])
            self._ensure_size(index, n_values)
            self._buffers[index][:n_values] = value
            return n_values
    # helper to fill a buffer with a color value
    def _fill_buffer_with_colors(self, index, color):
        arr = np.require(color, requirements='C', dtype=np.float32).flatten()
        self._ensure_size(index, len(arr))
        self._buffers[index][:len(arr)] = arr
        return len(arr) / 3


    def fill_styled_circles(self, x, y, radius, color, alpha=1):
        self._buffers[5][0:5] = [
            self._fill_buffer_with_scalars(0, x),
            self._fill_buffer_with_scalars(1, y),
            self._fill_buffer_with_scalars(2, radius),
            self._fill_buffer_with_colors(3, color),
            self._fill_buffer_with_scalars(4, alpha),
        ]   
        self._ctx.fillStyledCircles(*self._js_buffers[:6])

    def stroke_styled_circles(self, x, y, radius, color, alpha=1):
        self._buffers[5][0:5] = [
            self._fill_buffer_with_scalars(0, x),
            self._fill_buffer_with_scalars(1, y),
            self._fill_buffer_with_scalars(2, radius),
            self._fill_buffer_with_colors(3, color),
            self._fill_buffer_with_scalars(4, alpha),
        ]   
        self._ctx.strokeStyledCircles(*self._js_buffers[:6])
    
    def fill_circles(self, x, y, radius):
        self._buffers[3][0:3] = [
            self._fill_buffer_with_scalars(0, x),
            self._fill_buffer_with_scalars(1, y),
            self._fill_buffer_with_scalars(2, radius),
        ]
        self._ctx.fillCircles(*self._js_buffers[:4])
    
    def stroke_circles(self, x, y, radius):
        self._buffers[3][0:3] = [
            self._fill_buffer_with_scalars(0, x),
            self._fill_buffer_with_scalars(1, y),
            self._fill_buffer_with_scalars(2, radius),
        ]
        self._ctx.strokeCircles(*self._js_buffers[:4])
    
    def fill_rects(self, x, y, width, height):
        self._buffers[4][0:4] = [
            self._fill_buffer_with_scalars(0, x),
            self._fill_buffer_with_scalars(1, y),
            self._fill_buffer_with_scalars(2, width),
            self._fill_buffer_with_scalars(3, height),
        ]
        self._ctx.fillRects(*self._js_buffers[:5])
    
    def stroke_rects(self, x, y, width, height):
        self._buffers[4][0:4] = [
            self._fill_buffer_with_scalars(0, x),
            self._fill_buffer_with_scalars(1, y),
            self._fill_buffer_with_scalars(2, width),
            self._fill_buffer_with_scalars(3, height),
        ]
        self._ctx.strokeRects(*self._js_buffers[:5])
    
    def fill_styled_rects(self, x, y, width, height, color, alpha=1):
        self._buffers[6][0:6] = [
            self._fill_buffer_with_scalars(0, x),
            self._fill_buffer_with_scalars(1, y),
            self._fill_buffer_with_scalars(2, width),
            self._fill_buffer_with_scalars(3, height),
            self._fill_buffer_with_colors(4, color),
            self._fill_buffer_with_scalars(5, alpha),
        ]
        self._ctx.fillStyledRects(*self._js_buffers[:7])
    
    def stroke_styled_rects(self, x, y, width, height, color, alpha=1):
        self._buffers[6][0:6] = [
            self._fill_buffer_with_scalars(0, x),
            self._fill_buffer_with_scalars(1, y),
            self._fill_buffer_with_scalars(2, width),
            self._fill_buffer_with_scalars(3, height),
            self._fill_buffer_with_colors(4, color),
            self._fill_buffer_with_scalars(5, alpha),
        ]
        self._ctx.strokeStyledRects(*self._js_buffers[:7])
    
    def fill_arcs(self, x, y, radius, start_angle, end_angle, anticlockwise=False):
        self._buffers[5][0:5] = [
            self._fill_buffer_with_scalars(0, x),
            self._fill_buffer_with_scalars(1, y),
            self._fill_buffer_with_scalars(2, radius),
            self._fill_buffer_with_scalars(3, start_angle),
            self._fill_buffer_with_scalars(4, end_angle),
        ]
        self._ctx.fillArcs(*self._js_buffers[:6], anticlockwise)
    
    def stroke_arcs(self, x, y, radius, start_angle, end_angle, anticlockwise=False):
        self._buffers[5][0:5] = [
            self._fill_buffer_with_scalars(0, x),
            self._fill_buffer_with_scalars(1, y),
            self._fill_buffer_with_scalars(2, radius),
            self._fill_buffer_with_scalars(3, start_angle),
            self._fill_buffer_with_scalars(4, end_angle),
        ]
        self._ctx.strokeArcs(*self._js_buffers[:6], anticlockwise)
    
    def fill_styled_arcs(self, x, y, radius, start_angle, end_angle, color, alpha=1, anticlockwise=False):
        self._buffers[7][0:7] = [
            self._fill_buffer_with_scalars(0, x),
            self._fill_buffer_with_scalars(1, y),
            self._fill_buffer_with_scalars(2, radius),
            self._fill_buffer_with_scalars(3, start_angle),
            self._fill_buffer_with_scalars(4, end_angle),
            self._fill_buffer_with_colors(5, color),
            self._fill_buffer_with_scalars(6, alpha),
        ]
        self._ctx.fillStyledArcs(*self._js_buffers[:8], anticlockwise)
    
    def stroke_styled_arcs(self, x, y, radius, start_angle, end_angle, color, alpha=1, anticlockwise=False):
        self._buffers[7][0:7] = [
            self._fill_buffer_with_scalars(0, x),
            self._fill_buffer_with_scalars(1, y),
            self._fill_buffer_with_scalars(2, radius),
            self._fill_buffer_with_scalars(3, start_angle),
            self._fill_buffer_with_scalars(4, end_angle),
            self._fill_buffer_with_colors(5, color),
            self._fill_buffer_with_scalars(6, alpha),
        ]
        self._ctx.strokeStyledArcs(*self._js_buffers[:8], anticlockwise)

    # for polygons / line segments with potentially different number of points per polygon / line segment
    def _prepare_multipoint(self, points, points_per_item=None):
        if isinstance(points, list):
            if points_per_item is not None:
                raise RuntimeError("when points are a list, points_per_item must be None")
            
            points_per_item = []
            np_polygons = []
            for i, polygon_points in enumerate(points):
                polygon_points = np.require(polygon_points, requirements=["C"])
                if polygon_points.shape[1] != 2:
                    raise RuntimeError(
                        f"item {i} in points have wrong shape: `{polygon_points.shape}` but must be of type (n,2)"
                    )
                points_per_item.append(polygon_points.shape[0])
                np_polygons.append(polygon_points.ravel())

            num_polygons = len(points)
            flat_points = np.concatenate(np_polygons)
            points_per_item = np.array(points_per_item)

            return flat_points, points_per_item, num_polygons
        else:
            raise RuntimeError("points must be a list of numpy arrays or a numpy array with shape (n,2)")

    def fill_polygons(self, points, points_per_polygon=None):
        flat_points, points_per_item, num_items = self._prepare_multipoint(points, points_per_polygon)
        self._buffers[2][0:2] = [
            self._fill_buffer_with_scalars(0, flat_points),
            self._fill_buffer_with_scalars(1, points_per_item),
        ]
        self._ctx.fillPolygons(num_items, *self._js_buffers[:3])
    
    def stroke_polygons(self, points, points_per_polygon=None):
        flat_points, points_per_item, num_items = self._prepare_multipoint(points, points_per_polygon)
        self._buffers[2][0:2] = [
            self._fill_buffer_with_scalars(0, flat_points),
            self._fill_buffer_with_scalars(1, points_per_item)
        ]
        self._ctx.strokePolygons(num_items, *self._js_buffers[:3])

    def fill_styled_polygons(self, points, color, alpha=1, points_per_polygon=None): 
        flat_points, points_per_item, num_items = self._prepare_multipoint(points, points_per_polygon)
        self._buffers[4][0:4] = [
            self._fill_buffer_with_scalars(0, flat_points),
            self._fill_buffer_with_scalars(1, points_per_item),
            self._fill_buffer_with_colors(2,  color),
            self._fill_buffer_with_scalars(3, alpha),
        ]
        self._ctx.fillStyledPolygons(num_items, *self._js_buffers[:5])
    
    def stroke_styled_polygons(self, points, color, alpha=1, points_per_polygon=None):
        flat_points, points_per_item, num_items = self._prepare_multipoint(points, points_per_polygon)
        self._buffers[4][0:4] = [
            self._fill_buffer_with_scalars(0, flat_points.shape[0]),
            self._fill_buffer_with_scalars(1, points_per_item.shape[0]),
            self._fill_buffer_with_colors(2,  color),
            self._fill_buffer_with_scalars(3, alpha),
        ]
        self._ctx.strokeStyledPolygons(num_items, *self._js_buffers[:5])

    # stroke_line_segments
    def stroke_line_segments(self, points, points_per_segment=None):
        flat_points, points_per_item, num_items = self._prepare_multipoint(points, points_per_segment)
        self._buffers[2][0:2] = [
            self._fill_buffer_with_scalars(0, flat_points),
            self._fill_buffer_with_scalars(1, points_per_item)
        ]
        self._ctx.strokeLineSegments(num_items, *self._js_buffers[:3])
    
    def stroke_styled_line_segments(self, points, color, alpha=1, points_per_segment=None):
        flat_points, points_per_item, num_items = self._prepare_multipoint(points, points_per_segment)
        self._buffers[4][0:4] = [
            self._fill_buffer_with_scalars(0, flat_points),
            self._fill_buffer_with_scalars(1, points_per_item),
            self._fill_buffer_with_colors(2, color),
            self._fill_buffer_with_scalars(3, alpha),
        ]
        self._ctx.strokeStyledLineSegments(num_items, *self._js_buffers[:5])


# helper function to create a property that maps to a js property
def _make_prop(js_name):
    @property
    def prop(self):
        return getattr(self._ctx, js_name)
    @prop.setter
    def prop(self, value):
        setattr(self._ctx, js_name, value)

    return prop
    
# helper function st. we can have pythonic properties (ie `fill_style` instead of `fillStyle`)
def _extend_canvas():

    # add properties to the Canvas class
    py_to_js_name = {
        'fill_style': 'fillStyle',
        'stroke_style': 'strokeStyle',
        'global_alpha': 'globalAlpha',
        'font': 'font',
        'text_align': 'textAlign',
        'text_baseline': 'textBaseline',
        'direction': 'direction',
        'global_composite_operation': 'globalCompositeOperation',
        'shadow_offset_x': 'shadowOffsetX',
        'shadow_offset_y': 'shadowOffsetY',
        'shadow_blur': 'shadowBlur',
        'shadow_color': 'shadowColor',
        'line_width': 'lineWidth',
        'line_cap': 'lineCap',
        'line_join': 'lineJoin',
        'miter_limit': 'miterLimit',
        'filter': 'filter',
        'image_smoothing_enabled': 'imageSmoothingEnabled',
        'line_dash_offset': 'lineDashOffset'
    }
    for py_name, js_name in py_to_js_name.items():
        prop = _make_prop(js_name)
        setattr(OffscreenCanvas, py_name, prop)


_extend_canvas()
del _extend_canvas


from ipywidgets import DOMWidget
from traitlets import Int
from .._frontend import module_name, module_version

from traitlets import Unicode
import random
import string
import numpy as np
import pyjs
import asyncio
from pathlib import Path
import pyjs


# helper function to execute a javascript file.
def _exec_js_file(filename):
    try:
        with open(filename, "r") as f:
            js_code = f.read()
        
        pyjs.js.Function(js_code)()
    except Exception as e:
        raise RuntimeError(f"Error executing JavaScript file {filename}: {e}") from e

# execute the init.js file to initialize the js environment
def _init_js():
    THIS_DIR = Path(__file__).parent
    _exec_js_file(THIS_DIR/ "js" / "init.js")
_init_js()
del _init_js

# javascript object that contains (helper-) functions
# that are implemented in the init.js file.
_ipycanvas_js = pyjs.js.globalThis["_ipycanvas"]

# we store the canvas under a random name in the globalScope
# to avoid name clashes with other canvases.
def _rand_name():
    """Generate a random name for the canvas."""
    # Generate a random string of length 8
    return ''.join(random.choice(string.ascii_uppercase + string.digits) for _ in range(8))

# OffscreenCanvasCore contains "offscreen canvas" creating and event handling logic.
# But it **does not create** any drawing context.
# The 2d-drawing context is created in a derived class (ie OffscreenCanvas).
# This class (OffscreenCanvasCore) can also we used as a base class for other canvas types,
# For example a webgl / three.js canvas / or a pixi.js canvas.
class OffscreenCanvasCore(DOMWidget):
    """An offsceen canvas widget."""

    _model_name = Unicode('OffscreenCanvasModel').tag(sync=True)
    _model_module = Unicode(module_name).tag(sync=True)
    _model_module_version = Unicode(module_version).tag(sync=True)

    _view_name = Unicode('OffscreenCanvasView').tag(sync=True)
    _view_module = Unicode(module_name).tag(sync=True)
    _view_module_version = Unicode(module_version).tag(sync=True)

    # NOTE: the  _width and _height properties are only used to initialize sizes  
    # in the frontend. Since the canvas is then transfer to the worker thread,
    # we cannot change the size of the canvas anymore.
    _width = Int(300).tag(sync=True)
    _height = Int(150).tag(sync=True)
    _name = Unicode('_canvas_0').tag(sync=True)


    def __init__(self, width=300, height=150, *args, **kwargs):

        # once the canvas is displayed, we will store the javascript canvas object
        # in the _canvas attribute
        self._canvas = None 

        # create a random name for the canvas
        _name = _rand_name()

        # this "receiver" is used to pass events from the main-ui-thread to the worker thread.
        # it is a global object in the worker thread (ie where **this code is running**).
        self._receiver_name = f"_canvas_receiver_{_name}"

        # we store the name of the canvas in the _canvas_name as a global variable
        # in the worker thread (ie where this code is running).
        self._canvas_name = f"_canvas_{_name}"

        # helper function check if we already recived the canvas from the frontend.
        self._check_if_ready = pyjs.js.Function(f"""return "{self._canvas_name}" in globalThis""") 

        # we use this numpy arrray to store the mouse state
        # this is usefull, because we can access that array on the js side as typed array
        # and just write the values to it and read them here without any conversion.
        self.arr_mouse_state = np.array([0, 0, 0, 0], dtype=np.uint32)  # [is_inside, is_down, x, y]


        # in the frontend javascript code ** in the main-ui-thread** we will call a function
        # on a global object **in the worker thread**. (via comlink)
        # this global object is called "receiver" and is created in the worker thread.
        # This is used to pass events from the main-ui-thread to the worker thread.
        # see init.js for the implementation of "receiver_factory".
        self._js_receiver = _ipycanvas_js.receiver_factory(pyjs.buffer_to_js_typed_array(self.arr_mouse_state, view=True))
        pyjs.js.globalThis[self._receiver_name] =  self._js_receiver
        
 
        super().__init__(_name=_name, _width=width, _height=height, *args, **kwargs)


    def __del__(self):
        super().__del__()
        self._js_receiver.cleanup()
        pyjs.js.Function("receiver_name","""delete globalThis[receiver_name];""")(self._receiver_name)

    # getter setter for width and height
    @property
    def width(self):
        return self._width
    
    @width.setter
    def width(self, value):
        if self._canvas is not None:
            raise RuntimeError("OffscreenCanvasCore: Width can only be set before the canvas is displayed.")
        self._width = value
    
    @property
    def height(self):
        return self._height
    
    @height.setter
    def height(self, value):
        if self._canvas is not None:
            raise RuntimeError("OffscreenCanvasCore: Height can only be set before the canvas is displayed.")
        self._height = value


    # helper function to add a callback for an event an event
    def _on_event(self, event_name, callback):

        # for the offscreen canvas, we atm make the simplification that
        # only one callback can be set for each event.
        # This might change in the future, but for now it is sufficient.
        if self._js_receiver.has_property(f"on_{event_name}"):
            raise RuntimeError(f"Event '{event_name}' already has a callback set. Only one callback is allowed per event for the OffscreenCanvasCore.")

        # if the callback is a js function, we can just set it directly.
        if isinstance(callback, pyjs.JsValue):
            self._js_receiver[f"on_{event_name}"] = callback
        
        # if the callback is a python function, we need to create a js callable.
        # since this callback needs to be deleted later, we need to store the cleanup function
        else:
            js_callback, cleanup = pyjs.create_callable(callback)
            self._js_receiver[f"on_{event_name}"] = js_callback
            cleanup_js_fname = f"_cleanup_{event_name}"
            setattr(self._js_receiver, cleanup_js_fname, cleanup)
            self._js_receiver.add_to_cleanup(cleanup_js_fname)


    def on_mouse_enter(self, callback):
        self._on_event("mouse_enter", callback)
    
    def on_mouse_out(self, callback):
        self._on_event("mouse_leave", callback)

    def on_mouse_down(self, callback):
        self._on_event("mouse_down", callback)
    
    def on_mouse_up(self, callback):
        self._on_event("mouse_up", callback)
    
    def on_mouse_move(self, callback):
        self._on_event("mouse_move", callback)

    def on_key_down(self, callback):
        self._on_event("key_down", callback)

    def on_mouse_wheel(self, callback):
        self._on_event("mouse_wheel", callback)

    def on_key_up(self, callback):
        self._on_event("key_up", callback)

    def on_key_press(self, callback):
        self._on_event("key_press", callback)
    
    # touch events:
    # since ordinary canvas does not pass the id to the callbacks, we need
    # to make this the default behavior.
    def on_touch_start(self, callback, pass_id=False):
        if not pass_id:
            def wrapped(x, y, id=None):
                callback(x, y)
            self._on_event("touch_start", wrapped)
        else:
            self._on_event("touch_start", callback)
    def on_touch_end(self, callback, pass_id=False):
        if not pass_id:
            def wrapped(x, y, id=None):
                callback(x, y)
            self._on_event("touch_end", wrapped)
        else:
            self._on_event("touch_end", callback)

    def on_touch_move(self, callback, pass_id=False):
        if not pass_id:
            def wrapped(x, y, id=None):
                callback(x, y)
            self._on_event("touch_move", wrapped)
        else:
            self._on_event("touch_move", callback)

    def on_touch_cancel(self, callback, pass_id=False):
        if not pass_id:
            def wrapped(x, y, id=None):
                callback(x, y)
            self._on_event("touch_cancel", wrapped)
        else:
            self._on_event("touch_cancel", callback)
    

    def initialize(self):
        """ Initialize the canvas after it has been displayed.

            After the canvas has been displayed, we need to call this method to initialize the canvas.
            This method can only be used when the display method has been called in a **different cell**.

            # Cell1:
            ```python
            canvas = OffscreenCanvasCore(width=800, height=600)
            display(canvas)
            ```

            # Cell2:
            ```python
            canvas.initialize()
            # Now you can use the canvas
            # ...
            ```

            For more information, see the `async_initialize` method.
        """
        if not self._check_if_ready():
            raise RuntimeError(f"Canvas {self._canvas_name} is not ready. Call await canvas.adisplay() to display the canvas.") 
        self._canvas = pyjs.js.globalThis[self._canvas_name]
    
    async def async_initialize(self):
        """If we want to use the canvas in the same cell where it 
            was created **and displayed** we need to call this async function.
            While this sounds a bit counterintuitive, it is necessary because the
            offscreen canvas is created as regular canvas in the main-thread,
            and transfered to and recieved by the worker-thread. This transfering mechanism (in particular the
            receiving part) would be blocked by the cell execution.
            To get a chance to  receive the canvas in the worker-thread (ie where
            the kernel is running), we need to do run some asyc code (with some sleeping in between)

            ```python
            canvas = OffscreenCanvasCore(width=800, height=600)
            display(canvas)
            await canvas.async_initialize()
            # canvas is now ready to use
            # ...
        """

        
        c = 0
        while not  self._check_if_ready():
            await asyncio.sleep(0.1)
            c += 1
            if c >= 20:
                raise RuntimeError(f"Canvas {self._canvas_name} was not created in time.")
        
        self._canvas = pyjs.js.globalThis[self._canvas_name]

    def get_canvas(self):
        """Get the offscreen canvas javascript objectt."""
        if self._canvas is None:
            raise RuntimeError("Canvas is not displayed yet.")
        return self._canvas

    # mouse state
    def mouse_is_down(self):
        """Check if the mouse is currently pressed down."""
        return self.arr_mouse_state[1] == 1
    def mouse_is_inside(self):
        """Check if the mouse is currently inside the canvas."""
        return self.arr_mouse_state[0] == 1
    def mouse_position(self):
        """Get the current mouse position as a tuple (x, y)."""
        return (self.arr_mouse_state[2], self.arr_mouse_state[3])




"""
    this module provides a way to call a function repeatedly at a given frame rate.
    It is used to set a render loop for the canvas.
    It works with both offscreen canvas and regular canvas.

    When we are in a emscripten/wasm/lite environment, we can
    make use of the browsers requestAnimationFrame function.
    In a regular python environment, we use asyncio to call the function repeatedly
    as a task (we need to make it a task st. its not blocking the kernel,
    because otherwise we could not recieve events from the frontend).
"""

import time
import sys
from .canvas import hold_canvas as hold_classic_canvas

import sys
is_emscripten = sys.platform.startswith("emscripten")

# has pyjs
has_pyjs = True
try:
    import pyjs
except ImportError:
    has_pyjs = False

    
# if we are in a emscripten/wasm/lite environment, we can use the pyjs module
if has_pyjs and is_emscripten:
    import pyjs
    from .offscreen_canvas.offscreen_canvas_core import OffscreenCanvasCore

    # call a function repeatedly at a given frame rate
    # when fps is 0, requestAnimationFrame is used
    def call_repeated(func, fps=0):
        """Call a function repeatedly at a given frame rate.
        If fps is 0, requestAnimationFrame is used.

        Args:
            func: The function to call repeatedly.
            fps: The frame rate to call the function at. If 0, requestAnimationFrame
        """
        pyjs.set_main_loop_callback(func, fps)

    # set a render loop for the canvas
    # this is used to call the function repeatedly at a given frame rate
    # if the canvas is **not** an offscreen canvas, we use the hold_canvas context manager
    # st. we we only send one message to the frontend per frame
    def set_render_loop(canvas, func, fps=0):
        """Set a render loop for the canvas.
        This is used to call the function repeatedly at a given frame rate.
        If the canvas is **not** an offscreen canvas, we use the hold_canvas context
        manager st. we only send one message to the frontend per frame.

        Args:
            canvas: The canvas to set the render loop for.
            func: The function to call repeatedly.
            fps: The frame rate to call the function at. If 0, requestAnimationFrame
        """
        if isinstance(canvas, OffscreenCanvasCore):
            def wrapped_func(dt):
                try:
                    func(dt)
                except Exception as e:
                    # the best we can do there is catch the error and print it to the error stream
                    print(f"Error in requestAnimationFrame callback: {e}", file=sys.stderr)
                    pyjs.cancel_main_loop()
            pyjs.set_main_loop_callback(wrapped_func, 0)
        else:
            # For regular canvas we wrap eveything in a "hold_canvas" function
            def wrapped_func(dt):
                with hold_classic_canvas():
                    try:
                        func(dt)
                    except Exception as e:
                        # the best we can do there is catch the error and print it to the error stream
                        print(f"Error in requestAnimationFrame callback: {e}", file=sys.stderr)
                        pyjs.cancel_main_loop()
            pyjs.set_main_loop_callback(wrapped_func, fps)

        # return a lambda which can be used to cancel the loop
        return lambda: pyjs.cancel_main_loop()
else:
    import asyncio
    async def _call_repeated(func, fps):
        try:

            interval = 1 / fps
            last_start_time = time.time()

            while True:
                start_time = time.time()
                dt = start_time - last_start_time
                last_start_time = start_time
                try:
                    func(dt)
                except Exception as e:
                    print(f"Error in repeated function call: {e}", file=sys.stderr)
                    break


                elapsed_time = time.time() - start_time
                sleep_time = max(0, interval - elapsed_time)
                await asyncio.sleep(sleep_time)
        except asyncio.CancelledError:
            # If the task is cancelled, we just exit the loop
            pass
                
    
    def call_repeated(func, fps):
        """Call a function repeatedly at a given frame rate.
        Since we map an fps to requestAnimationFrame, for the 
        emscripten/lite environment, we use 60hz as default when fps is 0.

        Args:
            func: The function to call repeatedly.
            fps: The frame rate to call the function at. If 0, requestAnimationFrame
        """
        if fps == 0:
            # this is a special case, because for lite 
            # this mean "use requestAnimationFrame"
            # so here we just assume this means 60hz
            fps = 60

        loop = asyncio.get_event_loop()
        task = loop.create_task(_call_repeated(func, fps))

        # Return a lambda that can be used to cancel the loop
        return lambda: task.cancel()
        
    def set_render_loop(canvas, func, fps=0):
        """Set a render loop for the canvas.
        This is used to call the function repeatedly at a given frame rate.
        We use the hold_canvas context manager so we only send one message to the frontend per frame.

        Args:
            canvas: The canvas to set the render loop for.
            func: The function to call repeatedly.
            fps: The frame rate to call the function at. If 0, requestAnimationFrame
        """
        def wrapped_func(dt):
            with hold_classic_canvas():
                func(dt)
        return call_repeated(wrapped_func, fps)